import logging
import asyncio

from ..abc.source import Source

L = logging.getLogger(__name__)


class SubProcessSource(Source):
    """
    Sub-Process Source is capable of calling any command and fetch result from stdin as event

    Can be useful with commands like `tail -f`, `tshark -l -n -T ek -i wlan0` or others
    """

    ConfigDefaults = {
        "command": "",
        "line_len_limit": 2**20,
        "ok_return_codes": "0",  # Multiple codes must be separated by commas, e.g. 'ok_return_codes': '0,1,2',
    }

    def __init__(self, app, pipeline, *, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Command = str(self.Config["command"])
        assert self.Command, "`command` not set on " + self.__class__.__name__
        self._process = None
        # Get the list of OK return codes
        self.OKReturnCodes = [
            int(i.strip()) for i in self.Config["ok_return_codes"].split(",")
        ]

    async def main(self):
        self._process = await asyncio.create_subprocess_shell(
            self.Command,
            shell=True,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
            limit=self.Config.get("line_len_limit"),
        )
        # Check if process is running
        while self._process.returncode is None:
            await self.Pipeline.ready()
            event = await self._process.stdout.readline()
            await self.process(event)
        # Error message, when process has been terminated
        if (
            self._process.returncode not in self.OKReturnCodes
            and self._process.returncode is not None
        ):
            # Print error, wait a bit and retry again
            L.error(
                "Command {} has exited with return code: {}".format(
                    self.Command, self._process.returncode
                )
            )
            await asyncio.sleep(5)



import logging

import numpy as np

###

L = logging.getLogger(__name__)

###


class Index(object):
    def __init__(self, id=None):
        super().__init__()
        self.Id = id if id is not None else self.__class__.__name__

    def update(self, matrix):
        pass

    def serialize(self):
        return {
            "id": self.Id,
            "class": self.__class__.__name__,
        }

    def search(self, *args) -> set:
        raise NotImplementedError()


class BitMapIndex(Index):
    def __init__(self, column, matrix, id=None):
        """
        Make sure, that column values are discreet.
        Also the creation procedure is relatively slow.
        """
        super().__init__(id=id)
        self.Column = column
        self.BitMap = {}

        open_rows = list(matrix.I2NMap.keys())
        self.UniqueValues = list(np.unique(matrix.Array[self.Column][open_rows]))

        for u in self.UniqueValues:
            x = np.where(matrix.Array[self.Column] == u)
            self.BitMap[int(u)] = set(x[0].tolist())

    def search(self, value):
        """
        Returns set of matrix indexes.
        """

        return set(self.BitMap.get(str(value), []))

    def update(self, matrix):
        open_rows = list(matrix.I2NMap.keys())
        unique_values = set(np.unique(matrix.Array[self.Column][open_rows]))
        set_difference_left = (
            set(self.UniqueValues) - unique_values
        )  # there are some deleted vals
        set_difference_right = unique_values - set(
            self.UniqueValues
        )  # there are some added vals

        if len(set_difference_left) != 0:
            for set_member in set_difference_left:
                self.BitMap.pop(set_member)

        if len(set_difference_right) != 0:
            for set_member in set_difference_right:
                x = np.where(matrix.Array[self.Column] == set_member)
                self.BitMap[str(set_member)] = set(x[0].tolist())

        self.UniqueValues = list(unique_values)

    def serialize(self):
        serialized_bitmap = {}
        for key in self.BitMap:
            serialized_bitmap[key] = list(self.BitMap[key])

        serialized = super().serialize()
        serialized.update(
            {
                "bitmap": serialized_bitmap,
                "column": self.Column,
            }
        )

        return serialized

    def deserialize(self, data):
        self.BitMap = {}
        self.Column = data["column"]

        for key in data["bitmap"]:
            self.BitMap[key] = set(data["bitmap"][key])

        self.UniqueValues = list(self.BitMap.keys())


class TreeRangeIndex(Index):
    def __init__(self, column_start, column_end, matrix, id=None):
        super().__init__(id=id)
        self.ColumnStart = column_start
        self.ColumnEnd = column_end
        ranges = set()

        open_rows = list(matrix.I2NMap.keys())
        unique_start = np.unique(matrix.Array[column_start][open_rows])
        ranges |= set(unique_start)
        unique_end = np.unique(matrix.Array[column_end][open_rows])
        ranges |= set(unique_end)

        assert len(unique_start) == len(unique_end)  # ranges overlapping

        self.Ranges = sorted(list(ranges))

        self.MinValue = None
        self.MaxValue = None
        self.Tree = None

        if len(self.Ranges) != 0:
            self.MinValue = int(self.Ranges[0])
            self.MaxValue = int(self.Ranges[-1])

            self.Tree = self.sorted_array_to_bst(matrix, self.Ranges, [], [])

    def search(self, value):
        if self.Tree is None:
            return set()

        if (value < self.MinValue) or (value >= self.MaxValue):
            return set()

        subtree = self.Tree
        while True:
            node = subtree["node"]
            if node is None:
                return set(subtree["indexes"])

            if value < node:
                subtree = subtree["left"]
            else:
                subtree = subtree["right"]

        return set()

    def sorted_array_to_bst(self, matrix, arr, path, mask):
        if not arr:
            mask_ = mask + [False]
            if np.any(mask):
                path_ = path + [float("inf")]
                true_values = np.array(path_)[np.array(mask_)]
                lower = max(true_values)
                false_values = np.array(path_)[~np.array(mask_)]
                upper = min(false_values)
                r = tuple([lower, upper])
            else:
                path_ = path + [-float("inf")]
                r = tuple([path_[-1], path_[-2]])

            condition = (matrix.Array[self.ColumnEnd] <= r[1]) & (
                matrix.Array[self.ColumnStart] >= r[0]
            )
            indexes = np.where(condition)
            result = {
                "node": None,
                "indexes": indexes[0].tolist(),
                "left": None,
                "right": None,
            }

            return result

        mid = int(len(arr) / 2)
        root = dict(node=int(arr[mid]), indexes=[])
        root["left"] = self.sorted_array_to_bst(
            matrix, arr[:mid], path + [arr[mid]], mask + [False]
        )
        root["right"] = self.sorted_array_to_bst(
            matrix, arr[mid + 1 :], path + [arr[mid]], mask + [True]
        )
        return root

    def update(self, matrix):
        ranges = set()
        open_rows = list(matrix.I2NMap.keys())
        unique_start = np.unique(matrix.Array[self.ColumnStart][open_rows])
        ranges |= set(unique_start)
        unique_end = np.unique(matrix.Array[self.ColumnEnd][open_rows])
        ranges |= set(unique_end)

        # TODO: optimize this monster
        # if open_rows == self.OpenRows:
        # 	# partial update
        # 	# a) deleted values
        # 	# b) added values
        # 	# c) updated values, bug
        # else:
        # 	# full update

        assert len(unique_start) == len(unique_end)  # ranges overlapping
        self.Ranges = sorted(list(ranges))

        self.MinValue = None
        self.MaxValue = None
        self.Tree = None

        if len(self.Ranges) != 0:
            self.MinValue = self.Ranges[0]
            self.MaxValue = self.Ranges[-1]
            self.Tree = self.sorted_array_to_bst(matrix, self.Ranges, [], [])

    def serialize(self):
        serialized = super().serialize()
        ranges = []
        for r in self.Ranges:
            ranges.append(int(r))

        serialized.update(
            {
                "tree": self.Tree,
                "ranges": ranges,
                "column_start": self.ColumnStart,
                "column_end": self.ColumnEnd,
            }
        )
        return serialized

    def deserialize(self, data):
        self.Ranges = data["ranges"]
        if len(self.Ranges) != 0:
            self.MinValue = self.Ranges[0]
            self.MaxValue = self.Ranges[-1]
        self.Tree = data["tree"]
        self.ColumnStart = data["column_start"]
        self.ColumnEnd = data["column_end"]


class SliceIndex(Index):
    def __init__(self, column_start, column_end, matrix, resolution=None, id=None):
        super().__init__(id=id)
        self.ColumnStart = column_start
        self.ColumnEnd = column_end
        self.Resolution = resolution
        self.MinValue = None
        self.MaxValue = None
        self.SliceMap = None
        self._create_slices(matrix)

    def search(self, value):
        index = int((value - self.MinValue) // self.Resolution)
        key = self.MinValue + index * self.Resolution
        result = set(self.SliceMap.get(key, []))
        return result

    def update(self, matrix):
        self._create_slices(matrix)
        print(len(self.SliceMap), self.MinValue, self.MaxValue, self.Resolution)

    def _create_slices(self, matrix):
        if matrix.Array.shape[0] == 0:
            return

        self.SliceMap = {}
        open_rows = list(matrix.I2NMap.keys())
        self.MinValue = float(np.min(matrix.Array[self.ColumnStart][open_rows]))
        self.MaxValue = float(np.max(matrix.Array[self.ColumnEnd][open_rows]))

        if self.Resolution is None:
            diffs = (
                matrix.Array[self.ColumnEnd][open_rows]
                - matrix.Array[self.ColumnStart][open_rows]
            )
            self.Resolution = float(np.min(diffs))

        start_value = self.MinValue

        while start_value < self.MaxValue:
            end_value = start_value + self.Resolution
            condition = (matrix.Array[self.ColumnEnd] >= end_value) & (
                matrix.Array[self.ColumnStart] <= start_value
            )
            indexes = np.where(condition)
            self.SliceMap[start_value] = indexes[0].tolist()
            start_value = end_value

    def serialize(self):
        serialized = super().serialize()
        serialized.update(
            {
                "slice_map": self.SliceMap,
                "min_value": self.MinValue,
                "max_value": self.MaxValue,
                "resolution": self.Resolution,
                "column_start": self.ColumnStart,
                "column_end": self.ColumnEnd,
            }
        )
        return serialized

    def deserialize(self, data):
        # check how json works with float keys
        self.MinValue = data["min_value"]
        self.MaxValue = data["max_value"]
        self.Resolution = data["resolution"]
        self.SliceMap = {}
        for key in data["slice_map"]:
            self.SliceMap[float(key)] = data["slice_map"][key]

        self.Tree = data["tree"]
        self.ColumnStart = data["column_start"]
        self.ColumnEnd = data["column_end"]



from ...abc import SequenceExpression, evaluate


class TUPLE(SequenceExpression):
    Attributes = {
        "Items": [],
    }

    def __call__(self, context, event, *args, **kwargs):
        return tuple(
            evaluate(item, context, event, *args, **kwargs) for item in self.Items
        )



from netaddr import IPAddress

import netaddr.core

from bspump.declarative.abc import Expression, evaluate


class IP_FORMAT(Expression):
    """
    Parses IP address to its string representation.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Subnet": ["*"],  # TODO: This ...
    }

    Category = "IP"

    def __init__(self, app, *, arg_what, arg_format="auto"):
        super().__init__(app)
        self.Value = arg_what

        assert arg_format in frozenset("ipv4", "ipv6", "auto")
        self.Format = arg_format

    def __call__(self, context, event, *args, **kwargs):
        try:
            ip = IPAddress(evaluate(self.Value, context, event, *args, **kwargs))
        except netaddr.core.AddrFormatError:
            # IP address could not be detected
            return None

        if self.Format == "ipv6":
            return str(ip.ipv6())
        elif self.Format == "ipv4":
            return str(ip.ipv4())
        else:  # auto
            if (0xFFFF00000000 & ip.ipv6().value) == 0xFFFF00000000:
                return str(ip.ipv4())
            else:
                return str(ip.ipv6())



import asyncio
import email.mime.text
import logging

import aiosmtplib

from ..abc.connection import Connection

L = logging.getLogger(__name__)


class SmtpConnection(Connection):
    ConfigDefaults = {
        "server": "",
        "port": 587,
        "use_tls": False,
        "use_start_tls": False,
        "login": "",
        "password": "",
        "from": "",
        "to": "",
        "cc": "",
        "bcc": "",
        "subject": "Mail from ASAB",
        "output_queue_max_size": 10,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.Server = self.Config["server"]
        self.Port = self.Config["port"]
        self.UseTLS = self.Config.getboolean("use_tls")
        self.Use_STARTTLS = self.Config.getboolean("use_start_tls")

        self.Login = self.Config["login"]
        self.Password = self.Config["password"]

        self.From = self.Config["from"]
        self.To = self.Config["to"]
        self.Cc = self.Config["cc"]
        self.Bcc = self.Config["bcc"]

        self.Subject = self.Config["subject"]

        self.Smtp = None
        self.Loop = app.Loop
        self.PubSub = app.PubSub

        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self._output_queue = asyncio.Queue(maxsize=self._output_queue_max_size + 1)
        self.LoaderTask = asyncio.ensure_future(self._loader())

        self.PubSub.subscribe("Application.exit!", self._on_exit)

    async def _on_exit(self, event_name):
        # Wait till the _loader() terminates
        pending = [self.LoaderTask]
        while len(pending) > 0:
            # By sending None via queue, we signalize end of life
            await self._output_queue.put(None)
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )
            if self.Smtp is not None:
                self.Smtp.close()

    def consume(self, mail_message):
        self._output_queue.put_nowait(mail_message)

        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("SMTPConnection.pause!", self)

    async def _loader(self):
        loop = self.Loop

        self.Smtp = aiosmtplib.SMTP(
            hostname=self.Server, port=self.Port, loop=loop, use_tls=self.UseTLS
        )

        await self.Smtp.connect()

        if self.Use_STARTTLS is True:
            await self.Smtp.starttls()

        if self.Login != "" and self.Password != "":
            await self.Smtp.auth_login(self.Login, self.Password)

        while True:
            message_text = await self._output_queue.get()
            if message_text is None:
                break

            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.PubSub.publish(
                    "SMTPConnection.unpause!", self, asynchronously=True
                )

            message = email.mime.text.MIMEText(message_text)
            message["From"] = self.From
            message["To"] = self.To
            message["Subject"] = self.Subject
            if self.Cc != "":
                message["Cc"] = self.Cc
            if self.Bcc != "":
                message["Bcc"] = self.Bcc

            smtp_response, resp_text = await self.Smtp.send_message(message)
            # TODO: Not ideal way of the error detection, we need to investigate and refactor this.
            if resp_text[:2].lower() != "ok":
                L.error(f"Failed to send message: {resp_text}:{smtp_response}")



from bspump.http.web.components import BaseField
from bspump.http.web.template_env import env


class Button(BaseField):
    def __init__(self, button_id, name, on_click, **kwargs):
        super().__init__(name, **kwargs)
        self.name = name
        self.on_click = on_click
        self.button_id = button_id
        self.required: bool = kwargs.get("required", False)
        self.default_classes = kwargs.get(
            "default_css_classes",
            "bg-cyan-600 hover:bg-cyan-900 text-slate-100 font-bold py-2 px-4 rounded-lg",
        )

    def html(self, default="", readonly=False):
        template = env.get_template("button.html")
        return template.render(
            name=self.name,
            default_classes=self.default_classes,
            on_click=self.on_click,
            button_id=self.button_id,
        )



import contextlib
import logging
import ldap
import ldap.resiter

from ..abc.connection import Connection

#

L = logging.getLogger(__name__)

#


class LDAPObject(ldap.ldapobject.LDAPObject, ldap.resiter.ResultProcessor):
    pass


class LDAPConnection(Connection):
    """
    Examples of configurations:

    [connection:ldap]
    host=12.34.56.78
    port=27017
    username=cn=admin,dc=example,dc=org
    password=abc123def456

    [connection:ldap]
    uri=ldaps://localhost:636
    username=cn=admin,dc=example,dc=org
    password=abc123def456
    tls_cafile=/conf/cert.ca
    tls_require_cert=allow
    """

    ConfigDefaults = {
        "host": "localhost",
        "port": 0,  # = use the default 389 for non-secure and 636 for secure connection
        "uri": "",
        "username": "cn=admin,dc=example,dc=org",
        "password": "admin",
        # Path to CA file in PEM format
        "tls_cafile": "",
        # Certificate policy.
        # Possible options (from python-ldap docs):
        # "never"  - Don’t check server cert and host name
        # "allow"  - Used internally by slapd server.
        # "demand" - Validate peer cert chain and host name
        # "hard"   - Same as "demand"
        "tls_require_cert": "never",
        "network_timeout": "10",  # set network_timeout to -1 for no timeout
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)
        self.TLSEnabled = len(self.Config.get("tls_cafile")) > 0
        self.URI = self.Config.get("uri")
        self.Host = self.Config.get("host")
        self.Port = self.Config.getint("port")
        if self.Port == 0:
            self.Port = 636 if self.TLSEnabled else 389
        if len(self.URI) == 0:
            self.URI = "{scheme}://{host}:{port}".format(
                scheme="ldaps" if self.TLSEnabled else "ldap",
                host=self.Host,
                port=self.Port,
            )

    @contextlib.contextmanager
    def ldap_client(self):
        client = LDAPObject(self.URI)
        client.protocol_version = ldap.VERSION3
        client.set_option(ldap.OPT_REFERRALS, 0)
        client.set_option(
            ldap.OPT_NETWORK_TIMEOUT, int(self.Config.get("network_timeout"))
        )
        if self.TLSEnabled:
            self._enable_tls(client)

        try:
            client.simple_bind_s(
                self.Config.get("username"), self.Config.get("password")
            )
            yield client
        except Exception as e:
            L.error(
                "Cannot connect to LDAP server: {}".format(e),
                exc_info=True,
                struct_data={"ldap_uri": self.URI},
            )
        finally:
            client.unbind_s()

    def _enable_tls(self, client):
        tls_cafile = self.Config.get("tls_cafile")
        tls_require_cert = self.Config.get("tls_require_cert")
        if len(tls_cafile) > 0:
            client.set_option(ldap.OPT_X_TLS_CACERTFILE, tls_cafile)
        if tls_require_cert == "never":
            client.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        elif tls_require_cert == "demand":
            client.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_DEMAND)
        elif tls_require_cert == "allow":
            client.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_ALLOW)
        elif tls_require_cert == "hard":
            client.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_HARD)
        else:
            raise ValueError("Invalid 'tls_require_cert' value: {}.")
        client.set_option(ldap.OPT_X_TLS_NEWCTX, 0)



import numpy as np
import os


class WarmingUpCount(object):
    def __init__(self, size):
        self.DType = "i8"
        self.WUC = np.empty(size, dtype=self.DType)

    def decrease(self, indexes):
        self.WUC[indexes] -= 1
        self.WUC[self.WUC < 0] = 0

    def extend(self, size, value):
        start = self.WUC.shape[0]
        end = size
        self.WUC.resize(size, refcheck=False)
        self.WUC[start:end] = value

    def assign(self, index, value):
        self.WUC[index] = value

    def flush(self, indexes):
        self.WUC = self.WUC.take(indexes, axis=0)

    def create(self, size):
        pass

    def __len__(self):
        return self.WUC.shape[0]


class PersistentWarmingUpCount(WarmingUpCount):
    def __init__(self, path, size):
        super().__init__(size)
        self.Path = path
        if os.path.exists(self.Path):
            self.WUC = np.memmap(self.Path, dtype=self.DType, mode="readwrite")
        else:
            self.create(size)

    def extend(self, size, value):
        start = self.WUC.shape[0]
        end = size
        wuc = np.empty(self.WUC.shape[0], dtype=self.DType)
        wuc[:] = np.nan
        wuc[:] = self.WUC[:]
        wuc.resize(size, refcheck=False)
        self.WUC = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=wuc.shape)
        self.WUC[:] = wuc[:]
        self.WUC[start:end] = value

    def flush(self, indexes):
        super().flush(indexes)
        self.create(self.WUC.shape[0])

    def create(self, size):
        wuc = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=(size,))
        wuc[:] = self.WUC[:]
        self.WUC = wuc



import logging
import asyncio
import aiohttp
from ...abc.sink import Sink


L = logging.getLogger(__name__)


class HTTPClientWebSocketSink(Sink):
    ConfigDefaults = {
        "url": "https://localhost:8081/",
        "output_queue_max_size": 100,
    }

    def __init__(self, app, pipeline, ssl=None, id=None, config=None):
        super().__init__(app, pipeline, id, config)

        self.Queue = asyncio.Queue()

        self.QueueMaxSize = int(self.Config["output_queue_max_size"])
        assert self.QueueMaxSize >= 1

        self.SSL = ssl
        self.WebSocket = None
        self.Exiting = False

        self.Task = asyncio.ensure_future(self._task(self.Config["url"]))

        app.PubSub.subscribe("Application.exit!", self._on_exit)

        # We start throttled, till the connection to WS server is ready
        self.Pipeline.throttle(self, True)

    async def _task(self, url):
        async with aiohttp.ClientSession() as session:
            while not self.Exiting:
                try:
                    async with session.ws_connect(
                        url,
                        ssl=self.SSL,
                        heartbeat=30.0,
                    ) as ws:
                        self.WebSocket = ws
                        self.Pipeline.throttle(self, False)

                        try:
                            writer = asyncio.ensure_future(self._writer(ws))
                            reader = asyncio.ensure_future(self._reader(ws))
                            await asyncio.wait(
                                [writer, reader], return_when=asyncio.FIRST_COMPLETED
                            )

                        finally:
                            self.WebSocket = None
                            self.Pipeline.throttle(self, True)

                        if not ws.closed:
                            await ws.close()

                        if not writer.done():
                            writer.cancel()

                        try:
                            await writer
                        except asyncio.CancelledError:
                            pass

                        if not reader.done():
                            reader.cancel()

                        try:
                            await reader
                        except asyncio.CancelledError:
                            pass

                        if self.Exiting:
                            return

                except aiohttp.ClientError as e:
                    L.warning("WebSocket error: {}".format(e))

                except Exception as e:
                    L.warning("Error in websocket", exc_info=e)

                await asyncio.sleep(5)

    async def _writer(self, ws):
        while not ws.closed:
            event = await self.Queue.get()

            # Unthrottle if needed
            if self.Queue.qsize() == (self.QueueMaxSize - 1):
                self.Pipeline.throttle(self.Queue, False)

            if isinstance(event, (bytes, bytearray, memoryview)):
                await ws.send_bytes(event)
            else:
                await ws.send_str(event)

    async def _reader(self, ws):
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.ERROR:
                L.warning(
                    "WebSocket connection closed with exception {}".format(
                        ws.exception()
                    )
                )

            else:
                L.warning("WebSocket unknown/invalid message {}".format(msg))

        if not self.Exiting:
            L.warning("Peer closed a connection")

    async def _on_exit(self, event_name):
        self.Exiting = True

        if self.WebSocket is not None:
            await self.WebSocket.close()

        if self.Task is not None:
            t = self.Task
            self.Task = None
            await t

    def process(self, context, event):
        self.Queue.put_nowait(event)
        if self.Queue.qsize() == self.QueueMaxSize:
            self.Pipeline.throttle(self.Queue, True)



import unittest
import time

import bspump
import bspump.matrix
import bspump.unittest


class TestMatrix(bspump.unittest.TestCase):
    def test_matrix(self):
        matrix = bspump.Matrix(
            app=self.App,
            dtype=[
                ("f1", "i8"),
                ("f2", "i8"),
                ("f3", "i8"),
            ],
        )

        for i in range(100):
            n = matrix.add_row()
            matrix.Array[n][0] = 1  # Access by a field index
            matrix.Array[n]["f2"] = 1  # Access by a field name
            matrix.Array[n][2] = 1
            self.assertEqual(n, i)

        # closed = set()
        # closed |= matrix.ClosedRows.CR

        # for i in range(20, 40):
        # 	matrix.close_row(i)
        # 	closed.add(i)
        # 	self.assertIn(i, matrix.ClosedRows.CR)

        # self.assertEqual(closed, matrix.ClosedRows.CR)

        # for i in range(20):
        # 	n = matrix.add_row()
        # 	self.assertIn(n, closed)

    def test_matrix_zeros(self):
        matrix = bspump.Matrix(app=self.App)

        for i in range(100):
            n = matrix.add_row()
            self.assertEqual(n, i)

        matrix.zeros()

        self.assertEqual(matrix.Array.shape, (1,))

    # def test_matrix_flush(self):
    # 	matrix = bspump.Matrix(app=self.App)

    # 	for i in range(100):
    # 		n = matrix.add_row()
    # 		self.assertEqual(n, i)

    # 	for i in range(20, 40):
    # 		matrix.close_row(i)
    # 		self.assertIn(i, matrix.ClosedRows)

    # 	matrix.flush()

    # 	self.assertEqual(len(matrix.ClosedRows), 0)

    def test_matrix_dtypes(self):
        matrix = bspump.Matrix(
            app=self.App,
            dtype=[
                ("f1", "U20"),
                ("f2", "i8"),
                ("f3", "i8"),
            ],
        )

        row_index = matrix.add_row()
        row = matrix.Array[row_index]

        row["f1"] = "Ahoj"
        row["f2"] = 64



import bspump.unittest
from bspump.common import (
    Aggregator,
    StringAggregationStrategy,
    ListEventAggregationStrategy,
)


class TestAggregator(bspump.unittest.ProcessorTestCase):
    def test_aggregator(self):
        events = [
            ({}, {"message": "one"}),
            ({}, {"message": "two"}),
            ({}, {"message": "three"}),
            ({}, {"message": "four"}),
            ({}, {"message": "five"}),
            ({}, {"message": "six"}),
            ({}, {"message": "seven"}),
            ({}, {"message": "eight"}),
            ({}, {"message": "nine"}),
            ({}, {"message": "ten"}),
        ]

        self.set_up_processor(Aggregator, config={"completion_size": "5"})

        output = self.execute(events)

        self.assertEqual(2, len(output))

        self.assertEqual(
            (
                {},
                [
                    ({}, {"message": "one"}),
                    ({}, {"message": "two"}),
                    ({}, {"message": "three"}),
                    ({}, {"message": "four"}),
                    ({}, {"message": "five"}),
                ],
            ),
            output[0],
        )

        self.assertEqual(
            (
                {},
                [
                    ({}, {"message": "six"}),
                    ({}, {"message": "seven"}),
                    ({}, {"message": "eight"}),
                    ({}, {"message": "nine"}),
                    ({}, {"message": "ten"}),
                ],
            ),
            output[1],
        )

    def test_aggregator_flush_on_stop(self):
        events = [
            ({}, {"message": "one"}),
            ({}, {"message": "two"}),
            ({}, {"message": "three"}),
        ]

        self.set_up_processor(Aggregator, config={"completion_size": "2"})

        output = self.execute(events)

        self.assertEqual(2, len(output))

        self.assertEqual(
            ({}, [({}, {"message": "one"}), ({}, {"message": "two"})]), output[0]
        )

        self.assertEqual(({}, [({}, {"message": "three"})]), output[1])

    def test_list_event_aggragation_strategy(self):
        events = [
            ({}, {"message": "one"}),
            ({}, {"message": "two"}),
            ({}, {"message": "three"}),
            ({}, {"message": "four"}),
        ]

        self.set_up_processor(
            Aggregator,
            aggregation_strategy=ListEventAggregationStrategy(),
            config={"completion_size": "2"},
        )

        output = self.execute(events)

        self.assertEqual(2, len(output))

        self.assertEqual(({}, [{"message": "one"}, {"message": "two"}]), output[0])

        self.assertEqual(({}, [{"message": "three"}, {"message": "four"}]), output[1])

    def test_string_aggregation_strategy(self):
        events = [({}, "one"), ({}, "two"), ({}, "three")]

        self.set_up_processor(
            Aggregator,
            aggregation_strategy=StringAggregationStrategy(delimiter=";"),
            config={"completion_size": "3"},
        )

        output = self.execute(events)

        self.assertEqual(1, len(output))

        self.assertEqual(({}, "one;two;three"), output[0])



import ssl
from .config import Configurable


class SSLContextBuilder(Configurable):
    """
    Class for creating SSL context from a configuration.

    Examples:

    ```python
    ssl_context_builder = asab.tls.SSLContextBuilder(config_section)
    ssl_context = ssl_context_builder.build(protocol=ssl.PROTOCOL_TLS_CLIENT)
    # ssl_context is later used as a parameter when making HTTP requests
    ```
    """

    ConfigDefaults = {
        "cert": "",  # The certfile string must be the path to a PEM file containing the certificate as well as any number of CA certificates needed to establish the certificate’s authenticity.
        "key": "",  # The keyfile string, if present, must point to a file containing the private key in. Otherwise the private key will be taken from certfile as well.
        "key_password": "",
        # Following three options are fed into SSLContext.load_verify_locations(...)
        "cafile": "",
        "capath": "",
        "cadata": "",
        "ciphers": "",
        "dh_params": "",
        "verify_mode": "",  # empty or one of CERT_NONE, CERT_OPTIONAL or CERT_REQUIRED
        "check_hostname": "",
        "options": "",
    }

    def build(self, protocol=ssl.PROTOCOL_TLS) -> ssl.SSLContext:
        """
        Create SSL Context for the specified protocol.

        Allowed `protocol` values:

        - ssl.PROTOCOL_TLS
        - ssl.PROTOCOL_TLS_CLIENT: used for the client
        - ssl.PROTOCOL_TLS_SERVER: used for the server

        Args:
                protocol: TLS protocol used for the communication.
        """
        ctx = ssl.SSLContext(protocol=protocol)

        ctx.options |= ssl.OP_NO_SSLv2
        ctx.options |= ssl.OP_NO_SSLv3

        keyfile = self.Config.get("key")
        if len(keyfile) == 0:
            keyfile = None

        key_password = self.Config.get("key_password")
        if len(key_password) == 0:
            key_password = None

        cert = self.Config.get("cert")
        if len(cert) != 0:
            ctx.load_cert_chain(
                cert,
                keyfile=keyfile,
                password=key_password,
            )

        cafile = self.Config.get("cafile")
        if len(cafile) == 0:
            cafile = None

        capath = self.Config.get("capath")
        if len(capath) == 0:
            capath = None

        cadata = self.Config.get("cadata")
        if len(cadata) == 0:
            cadata = None

        if (cafile is not None) or (capath is not None) or (cadata is not None):
            ctx.load_verify_locations(cafile=cafile, capath=capath, cadata=cadata)

        ciphers = self.Config.get("ciphers")
        if len(ciphers) != 0:
            ctx.set_ciphers(ciphers)

        dh_params = self.Config.get("dh_params")
        if len(dh_params) != 0:
            ctx.load_dh_params(dh_params)

        verify_mode = self.Config.get("verify_mode")
        if len(verify_mode) > 0:
            verify_mode_tx = {
                "CERT_NONE": ssl.CERT_NONE,
                "CERT_OPTIONAL": ssl.CERT_OPTIONAL,
                "CERT_REQUIRED": ssl.CERT_REQUIRED,
            }.get(verify_mode.upper())
            if verify_mode_tx is None:
                raise RuntimeError("Unknown value {}".format(verify_mode))
            ctx.verify_mode = verify_mode_tx

        # TODO: check_hostname > ctx.check_hostname
        # TODO: options > ctx.options

        return ctx



import logging

from ...abc import Expression
from ...declerror import DeclarationError

###

L = logging.getLogger(__name__)

###


class CUT(Expression):
    Attributes = {
        "Value": ["str"],
        "Delimiter": ["str"],
        "Field": ["*"],  # TODO: This ...
    }

    Category = "String"

    def __init__(self, app, *, arg_what, arg_delimiter, arg_field):
        super().__init__(app)

        self.Value = arg_what

        # TODO: Delimiter must be a single character string
        self.Delimiter = arg_delimiter

        # Must be an integer
        self.Field = arg_field

    def get_outlet_type(self):
        return str.__name__

    def consult_inlet_type(self, key, child):
        return str.__name__

    def __call__(self, context, event, *args, **kwargs):
        try:
            value = self.Value(context, event, *args, **kwargs)
            x = value.split(self.Delimiter, self.Field + 1)
            return x[self.Field]

        except Exception as e:
            L.exception(
                "The following exception ocurred in !CUT expression [delimiter: {}, field: {}]".format(
                    self.Delimiter, self.Field
                )
            )
            raise DeclarationError(original_exception=e, location=self.get_location())



import re

from ...abc import Expression
from ..value.valueexpr import VALUE


class REGEX(Expression):
    """
    Checks if `value` matches with the `regex`, returns `hit` / `miss` respectively.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Regex": ["*"],  # TODO: This ...
        "Hit": ["*"],
        "Miss": ["*"],
    }

    Category = "Regex"

    def __init__(self, app, *, arg_regex, arg_what, arg_hit=True, arg_miss=False):
        super().__init__(app)
        self.Value = arg_what
        self.Regex = re.compile(arg_regex)

        if not isinstance(arg_hit, Expression):
            self.Hit = VALUE(app, value=arg_hit)
        else:
            self.Hit = arg_hit

        if not isinstance(arg_miss, Expression):
            self.Miss = VALUE(app, value=arg_miss)
        else:
            self.Miss = arg_miss

        # TODO: Regex flags

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        match = re.search(self.Regex, value)
        if match is None:
            return self.Miss(context, event, *args, **kwargs)
        else:
            return self.Hit(context, event, *args, **kwargs)


class REGEX_PARSE(Expression):
    """
    Search `value` forr `regex` with regular expressions groups.
    If nothing is found `miss` is returned.
    Otherwise, groups are returned in as a list.

    If `fields` are provided, the groups are mapped to provided fields.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Regex": ["*"],  # TODO: This ...
        "Unset": [],
        "Update": ["*"],
        "Items": [],
    }

    Category = "Regex"

    def __init__(
        self,
        app,
        *,
        arg_regex,
        arg_what,
        arg_items=None,
        arg_miss=None,
        arg_set=None,
        arg_unset=None,
        arg_update=None
    ):
        super().__init__(app)

        if not isinstance(arg_what, Expression):
            self.Value = VALUE(app, value=arg_what)
        else:
            self.Value = arg_what

        self.Regex = re.compile(arg_regex)
        self.Items = arg_items

        if not isinstance(arg_miss, Expression):
            self.Miss = VALUE(app, value=arg_miss)
        else:
            self.Miss = arg_miss

        self.ArgSet = arg_set
        self.Set = None

        if arg_unset is not None:
            assert isinstance(arg_unset, list)
        self.Unset = arg_unset

        self.Update = arg_update

        # TODO: Regex flags

    def set(self, key, value):
        setattr(self, key, value)

        if "Set" in key:
            self.Set[key[3:]] = value

    def initialize(self):
        if self.ArgSet is None:
            self.Set = None
        else:
            assert isinstance(self.ArgSet, dict)
            self.Set = dict()
            self._set_value_or_expression_to_attribute(self.ArgSet, self.Set, "Set")

    def _set_value_or_expression_to_attribute(self, _from, _to, name):
        for key, value in _from.items():
            # Parse value to attribute, if not expression
            if isinstance(value, Expression):
                _to[key] = value
            else:
                assert (
                    isinstance(value, (int, str, bytes, bool, tuple, list))
                    or value is None
                )
                value = VALUE(self.App, value=value)
                _to[key] = value

            # Set the value to the class attribute
            attr_name = "{}{}".format(name, key)
            setattr(self, attr_name, value)
            self.Attributes[attr_name] = value.get_outlet_type()

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        try:
            match = re.search(self.Regex, value)
        except TypeError:
            match = None
        if match is None:
            return self.Miss(context, event, *args, **kwargs)

        groups = match.groups()
        if self.Items is None:
            return groups

        ret = dict()
        for item, group in zip(self.Items, groups):
            if isinstance(item, Expression):
                result = item(context, event, group, *args, **kwargs)
                if result is None:
                    return self.Miss(context, event, *args, **kwargs)
                ret[result] = group

            elif isinstance(item, dict):
                if group is None:
                    continue
                assert len(item) == 1

                key, value = next(iter(item.items()))

                if isinstance(value, Expression):
                    v = value(context, event, group, *args, **kwargs)
                    if v is not None:
                        ret[key] = v

                elif isinstance(value, list):
                    for valuei in value:
                        v = valuei(context, event, group, *args, **kwargs)
                        if v is not None:
                            ret[key] = v
                            break

                else:
                    raise RuntimeError("Unexpected type: '{}'".format(value))
            else:
                ret[item] = group

        if self.Set is not None:
            for key, value in self.Set.items():
                v = value(context, event, ret, *args, **kwargs)
                if v is not None:
                    ret[key] = v

        if self.Update is not None:
            update_dict = self.Update(context, event, ret, *args, **kwargs)
            if update_dict is not None and update_dict is not False:
                ret.update(update_dict)

        if self.Unset is not None:
            for key in self.Unset:
                ret.pop(key, None)

        return ret


class REGEX_REPLACE(Expression):
    """
    Search `regex` in `value` and replace with `replace`.

    See Python documentation of `re.sub()` for more details.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Regex": ["*"],  # TODO: This ...
        "Replace": ["*"],
    }

    Category = "Regex"

    def __init__(self, app, *, arg_regex, arg_replace, arg_what):
        super().__init__(app)
        self.Value = arg_what
        self.Regex = re.compile(arg_regex)

        if not isinstance(arg_replace, Expression):
            self.Replace = VALUE(app, value=arg_replace)
        else:
            self.Replace = arg_replace

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        repl = self.Replace(context, event, *args, **kwargs)
        return self.Regex.sub(repl, value)


class REGEX_SPLIT(Expression):
    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Regex": ["*"],  # TODO: This ...
        "Max": ["*"],
    }

    Category = "Regex"

    def __init__(self, app, *, arg_regex, arg_what, arg_max=0):
        super().__init__(app)
        self.Value = arg_what
        self.Regex = re.compile(arg_regex)

        if not isinstance(arg_max, Expression):
            self.Max = VALUE(app, value=arg_max)
        else:
            self.Max = arg_max

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        maxsplit = self.Max(context, event, *args, **kwargs)
        return self.Regex.split(value, maxsplit=maxsplit)


class REGEX_FINDALL(Expression):
    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Regex": ["*"],  # TODO: This ...
    }

    Category = "Regex"

    def __init__(self, app, *, arg_regex, arg_what):
        super().__init__(app)
        self.Value = arg_what
        self.Regex = re.compile(arg_regex)

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        return self.Regex.findall(value)



import unittest
import numpy as np

import bspump
import bspump.matrix
import bspump.unittest


class TestSessionMatrix(bspump.unittest.TestCase):
    def test_matrix_store(self):
        dtype = [
            ("f1", "U20"),
            ("f2", "i8"),
        ]
        matrix = bspump.matrix.SessionMatrix(app=self.App, dtype=dtype)
        n = 100
        for i in range(n):
            name = "id_" + str(i)
            index = matrix.add_row(str(name))
            row = np.zeros(1, dtype=dtype)
            f1 = "something" + str(i)
            row["f1"] = f1
            row["f2"] = i
            matrix.store(name, row)
            self.assertEqual(matrix.Array[index]["f1"], f1)
            self.assertEqual(matrix.Array[index]["f2"], i)

    def test_matrix_store_event(self):
        dtype = [("f1", "U20"), ("f2", "i8"), ("f3", "f8"), ("f4", "U30")]
        matrix = bspump.matrix.SessionMatrix(app=self.App, dtype=dtype)
        index = matrix.add_row("abc")
        event = {"id": "abc", "f0": 7, "f1": "abcd", "f2": 123, "f3": 123.0}
        matrix.store_event(index, event)
        row = np.zeros(1, dtype=dtype)
        row["f1"] = event["f1"]
        row["f2"] = event["f2"]
        row["f3"] = event["f3"]
        row["f4"] = ""

        for key in row.dtype.names:
            self.assertEqual(row[key], matrix.Array[index][key])

        index = matrix.add_row("bcd")
        event = {"id": "bcd", "f0": 666, "f2": 88, "f4": "cdef"}

        matrix.store_event(index, event, keys=["f2", "f4", "f5"])
        row = np.zeros(1, dtype=dtype)
        row["f1"] = ""
        row["f2"] = event["f2"]
        row["f3"] = 0
        row["f4"] = event["f4"]

        for key in row.dtype.names:
            self.assertEqual(row[key], matrix.Array[index][key])

    def test_matrix_decode_row(self):
        dtype = [
            ("f0", "U3"),
            ("f1", "U8"),
        ]
        matrix = bspump.matrix.SessionMatrix(
            app=self.App, dtype=dtype, config={"primary_name": "event_id"}
        )
        event_id = "abc"
        index = matrix.add_row(event_id)
        matrix.Array[index]["f0"] = "aaa"
        matrix.Array[index]["f1"] = "bbb"

        event = matrix.decode_row(index)
        self.assertEqual(len(event), 3)
        self.assertEqual(event["event_id"], event_id)
        self.assertEqual(event["f0"], matrix.Array[index]["f0"])
        self.assertEqual(event["f1"], matrix.Array[index]["f1"])

        event_id = "bcd"
        index = matrix.add_row(event_id)
        matrix.Array[index]["f0"] = "bbb"
        matrix.Array[index]["f1"] = "ccc"

        event = matrix.decode_row(index, keys=["f0"])
        self.assertEqual(len(event), 2)
        self.assertEqual(event["event_id"], event_id)
        self.assertEqual(event["f0"], matrix.Array[index]["f0"])



import unittest
import time
import numpy as np

import bspump
import bspump.matrix
import bspump.unittest


class TestGeoMatrix(bspump.unittest.TestCase):
    def test_matrix(self):
        bbox = {
            "min_lon": 14.259097,
            "max_lon": 14.589601,
            "min_lat": 49.974702,
            "max_lat": 50.160150,
        }
        matrix = bspump.matrix.GeoMatrix(app=self.App, bbox=bbox, resolution=5)
        self.assertEqual(
            int(matrix.MapHeight * matrix.Resolution), int(matrix.SizeHeight)
        )
        self.assertEqual(
            int(matrix.MapWidth * matrix.Resolution), int(matrix.SizeWidth)
        )
        self.assertEqual(matrix.Array.shape, (matrix.MapHeight, matrix.MapWidth))

    def test_matrix_is_in_boundaries(self):
        bbox = {
            "min_lon": 14.259097,
            "max_lon": 14.589601,
            "min_lat": 49.974702,
            "max_lat": 50.160150,
        }
        matrix = bspump.matrix.GeoMatrix(app=self.App, bbox=bbox, resolution=5)
        coordinates = [(0, 0), (0, 50), (14.3, 51), (14.3, 50), (13, 50)]
        ground_truths = [False, False, False, True, False]

        for i in range(len(coordinates)):
            lat = coordinates[i][1]
            lon = coordinates[i][0]
            ground_truth = ground_truths[i]
            matrix_answer = matrix.is_in_boundaries(lat, lon)
            self.assertIs(matrix_answer, ground_truth)

    def test_matrix_equirectangular(self):
        bbox = {
            "min_lon": 14.259097,
            "max_lon": 14.589601,
            "min_lat": 49.974702,
            "max_lat": 50.160150,
        }
        matrix = bspump.matrix.GeoMatrix(app=self.App, bbox=bbox, resolution=5)

        indexes = [
            (0, 0),
            (0, np.ceil(matrix.MapWidth / 2)),
            (0, matrix.MapWidth - 1),
            ((np.ceil(matrix.MapHeight / 2)), 0),
            (matrix.MapHeight - 1, matrix.MapWidth - 1),
        ]

        for i in range(len(indexes)):
            row = indexes[i][0]
            column = indexes[i][1]
            lat, lon = matrix.inverse_equirectangular(row, column)
            row_, column_ = matrix.project_equirectangular(lat, lon)
            self.assertEqual(row_, row)
            self.assertEqual(column_, column)



import numpy as np
import os
import collections


class Index(object):
    def __init__(self):
        self.N2IMap = collections.OrderedDict()
        self.I2NMap = collections.OrderedDict()

    def pop_index(self, index):
        if index in self.I2NMap:
            row_name = self.I2NMap.pop(index)
            del self.N2IMap[row_name]
            return True
        return False

    def get_row_index(self, row_name):
        return self.N2IMap.get(row_name)

    def get_row_name(self, row_index):
        return self.I2NMap.get(row_index)

    def add_row(self, name, index):
        self.N2IMap[name] = index
        self.I2NMap[index] = name

    def flush(self, indexes):
        n2imap = collections.OrderedDict()
        i2nmap = collections.OrderedDict()
        i = 0
        saved_indexes = []
        for row_name, row_index in self.N2IMap.items():
            if row_index not in indexes:
                n2imap[row_name] = i
                i2nmap[i] = row_name
                saved_indexes.append(row_index)
                i += 1

        self.N2IMap = n2imap
        self.I2NMap = i2nmap
        return saved_indexes

    def serialize(self):
        return {
            "N2IMap": self.N2IMap,
            "I2NMap": self.I2NMap,
        }

    def deserialize(self, data):
        self.N2IMap = data["N2IMap"]
        self.I2NMap = data["I2NMap"]

    def extend(self, size):
        pass

    def __contains__(self, row_name):
        if row_name in self.N2IMap:
            return True

        return False

    def __len__(self):
        return len(self.N2IMap)


class PersistentIndex(Index):
    def __init__(self, path, size):
        super().__init__()
        self.DType = "U30"
        self.Path = path

        if os.path.exists(self.Path):
            self.Map = np.memmap(self.Path, dtype=self.DType, mode="readwrite")
            for i in range(len(self.Map)):
                value = self.Map[i]
                if value != "":
                    self.I2NMap[i] = value
                    self.N2IMap[value] = i
        else:
            if size is None:
                raise RuntimeError("The size should correspond to array size")

            self.Map = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=(size,))

    def pop_index(self, index):
        if super().pop_index(index):
            self.Map[index] = ""
            return True
        return False

    def add_row(self, name, index):
        super().add_row(name, index)
        self.Map[index] = name

    def extend(self, size):
        map_ = np.zeros(self.Map.shape[0], dtype=self.DType)
        map_[:] = self.Map[:]
        map_.resize(size, refcheck=False)
        self.Map = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=map_.shape)
        self.Map[:] = map_[:]

    def flush(self, closed_indexes):
        saved_indexes = super().flush(closed_indexes)
        self.Map = self.Map.take(saved_indexes, axis=0)
        map_ = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=self.Map.shape)
        map_[:] = self.Map[:]
        self.Map = map_
        return saved_indexes



from .trigger import Trigger


class OpportunisticTrigger(Trigger):
    """
    This trigger tries to trigger the pump as frequenty as possible.
    It triggers immediatelly when possible, after each Source report completed cycle and in 5 sec. period (see chilldown period)
    """

    def __init__(self, app, id=None, run_immediately=True, chilldown_period=5):
        super().__init__(app, id=id)
        self.ChilldownPeriod = chilldown_period  # Seconds

        app.PubSub.subscribe("Application.tick!", self.on_tick)

        if run_immediately:
            self.Loop.call_soon(self.on_tick)

    def pause(self, pause=True):
        super().pause(pause)
        if not self.Paused:
            self.Loop.call_soon(self.on_tick)

    def on_tick(self, event_type="simulated"):
        now = self.Loop.time()
        if (self.LastFireAt != 0) and (now < (self.LastFireAt + self.ChilldownPeriod)):
            return

        self.fire()

    def done(self, trigger_source):
        self.Loop.call_soon(self.on_tick)

    @classmethod
    def construct(cls, app, definition: dict):
        id = definition.get("id")
        run_immediately = definition.get("args", {}).get("run_immediately", True)
        chilldown_period = definition.get("args", {}).get("chilldown_period", 5)
        return cls(
            app,
            id=id,
            run_immediately=run_immediately,
            chilldown_period=chilldown_period,
        )



from .processor import ProcessorBase


class Sink(ProcessorBase):
    """
    Sink is basically a processor. It takes an event sends it to a database where it is stored.

    |

    """

    def handle_error(self, context, event, exception, timestamp):
        raise exception



import sys
import pprint
from ..abc.sink import Sink
from ..abc.processor import Processor


class PrintSink(Sink):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        """
        print(event, file=self.Stream)


class PPrintSink(Sink):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        """
        pprint.pprint(event, stream=self.Stream)


class PrintProcessor(Processor):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        :return: event
        """
        print(event, file=self.Stream)
        return event


class PPrintProcessor(Processor):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        :return: event
        """
        pprint.pprint(event, stream=self.Stream)
        return event


class PrintContextProcessor(Processor):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        :return: event

        """

        print(context, file=self.Stream)
        return event


class PPrintContextProcessor(Processor):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None, stream=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Stream = stream if stream is not None else sys.stdout

    def process(self, context, event):
        """
        Description:

        |

        :return: event
        """
        pprint.pprint(context, stream=self.Stream)
        return event



import re

# HOW TO FULLFIL OPEMETRICS STANDARD

# Metrics SHOULD have "unit" and "help" Tags
# Help is a string and SHOULD be non-empty. It is used to give a brief description of the MetricFamily for human consumption and SHOULD be short enough to be used as a tooltip.
# Values MUST be float or integer. Boolean values MUST follow 1==true, 0==false.
# Colons in MetricFamily names are RESERVED to signal that the MetricFamily is the result of a calculation or aggregation of a general purpose monitoring system. MetricFamily names beginning with underscores are RESERVED and MUST NOT be used unless specified by OpenMetric standard. - Anything that is not A-Z, a-z or digit is transformed into "_" and leading "_" is stripped.
# NaN is a number like any other in OpenMetrics, usually resulting from a division by zero such as for a summary quantile if there have been no observations recently. NaN does not have any special meaning in OpenMetrics, and in particular MUST NOT be used as a marker for missing or otherwise bad data.

# Feel free to read more about OpenMetrics standard here: https://github.com/OpenObservability/OpenMetrics/blob/main/specification/OpenMetrics.md


def metric_to_openmetric(m):
    metric_lines = []
    if m.get("type") in ["Histogram", "HistogramWithDynamicTags"]:
        metric_type = "histogram"
    elif (
        m.get("type")
        in ["Counter", "CounterWithDynamicTags", "AggregationCounterWithDynamicTags"]
        and m.get("reset") is False
    ):
        metric_type = "counter"
    else:
        metric_type = "gauge"
    m_name = m.get("name")
    unit = m.get("unit")
    if unit:
        unit = validate_format(unit)
    name = get_full_name(m_name, unit)
    help = m.get("help")
    fieldset = m.get("fieldset")

    metric_lines.append(translate_metadata(name, metric_type, unit, help))

    if metric_type == "histogram":
        for field in fieldset:
            if m.get("reset") is False:
                values = field.get("actuals")
            else:
                values = field.get("values")

            # SKIP empty fields
            if all([bucket == {} for bucket in values.get("buckets").values()]):
                continue

            for upperbound, bucket in values.get("buckets").items():
                if bucket == {}:
                    continue
                for v_name, value in bucket.items():
                    histogram_labels = field.get("tags").copy()
                    histogram_labels.update({"le": str(upperbound)})
                    if validate_value(value) is False:
                        continue
                    metric_lines.append(
                        translate_value(
                            name, v_name, value, metric_type, histogram_labels
                        )
                    )
            metric_lines.append(
                translate_value(
                    name + "_count",
                    None,
                    values.get("count"),
                    metric_type,
                    field.get("tags"),
                )
            )
            metric_lines.append(
                translate_value(
                    name + "_sum",
                    None,
                    values.get("sum"),
                    metric_type,
                    field.get("tags"),
                )
            )

    else:
        for field in fieldset:
            if metric_type == "counter":
                values = field.get("actuals")
            elif m.get("type") == "AggregationCounter" and m.get("reset") is False:
                values = field.get("actuals")
            else:
                values = field.get("values")
            for v_name, value in values.items():
                if validate_value(value) is False:
                    continue
                metric_lines.append(
                    translate_value(name, v_name, value, metric_type, field.get("tags"))
                )

    metric_text = "\n".join(metric_lines)
    return metric_text


def validate_format(name):
    name = str(name)
    regex = r"[a-zA-Z:][a-zA-Z0-9_:]*"
    match = re.fullmatch(regex, name)
    if match is None:
        regex_sub = r"[^a-zA-Z0-9_:]"
        name = re.sub(regex_sub, "_", name)
        name = name.lstrip("_0123456789")
    return name


def get_full_name(m_name, unit):
    name = validate_format(m_name)
    if unit:
        name += "_{}".format(unit)
    return name


def get_tags_labels(tags):
    labels_dict = {}
    for tag, tag_v in tags.items():
        if tag in {"unit", "help"}:
            continue
        else:
            labels_dict[validate_format(tag)] = tag_v
    return labels_dict


def translate_metadata(name, type, unit, help):
    meta_lines = []
    meta_lines.append("# TYPE {} {}".format(name, type))
    if unit:
        meta_lines.append("# UNIT {} {}".format(name, unit))
    if help:
        meta_lines.append("# HELP {} {}".format(name, help))
    metadata = "\n".join(meta_lines)
    return metadata


def validate_value(value):
    f = isinstance(value, (int, float))
    # isinstance(True, int) => True -> bool is instance of int
    if type(value) is bool:
        f = False
    return f


def translate_value(name, v_name, value, metric_type, labels_dict):
    labels_dict = {validate_format(k): v for k, v in labels_dict.items()}
    labels_str = get_value_labels(v_name, labels_dict)

    if metric_type == "counter":
        name = name + "_total"

    return "{}{} {}".format(name, labels_str, value)


def get_value_labels(v_name, labels_dict):
    if v_name is not None:
        labels_dict.update({"name": v_name})

    labels_str = "{"
    if labels_dict != {}:
        labels_str += ",".join(['{}="{}"'.format(k, v) for k, v in labels_dict.items()])
    labels_str += "}"
    if len(labels_str) <= 2:
        return None
    return labels_str



import logging

from ..abc.source import TriggerSource


L = logging.getLogger(__name__)


class ODBCSource(TriggerSource):
    ConfigDefaults = {"query": ""}

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop
        self._connection = pipeline.locate_connection(app, connection)
        self._query = self.Config["query"]
        self.App = app

    async def cycle(self):
        await self._connection.ConnectionEvent.wait()
        async with self._connection.acquire() as connection:
            async with connection.cursor() as cur:
                await cur.execute(self._query)
                event = {}
                while True:
                    await self.Pipeline.ready()
                    row = await cur.fetchone()
                    if row is None:
                        break

                    # Transform row to an event object
                    for i, val in enumerate(row):
                        event[cur.description[i][0]] = val

                    # Pass event to the pipeline
                    await self.process(event)



import logging

from ..abc.sink import Sink
from .connection import ElasticSearchBulk

from .data_feeder import data_feeder_create_or_index

#

L = logging.getLogger(__name__)

#


class ElasticSearchSink(Sink):
    """
    ElasticSearchSink allows you to insert events into ElasticSearch through POST requests

    The following attributes can be passed to the context and thus override the default behavior
    of the sink:

    es_index (STRING): ElasticSearch index name

    data_feeder accepts the event as its only parameter and yields data as Python generator
    The example implementation is:

    def data_feeder_create_or_index(event):
                    _id = event.pop("_id", None)

                    if _id is None:
                                    yield b'{"create":{}}\n'
                    else:
                                    yield orjson.dumps(
                                                    {"index": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE
                                    )

                    yield orjson.dumps(event, option=orjson.OPT_APPEND_NEWLINE)

    |

    """

    ConfigDefaults = {
        "index_prefix": "bspump_",  # Obsolete, use 'index'
        "index": "bspump_",
    }

    def __init__(
        self,
        app,
        pipeline,
        connection,
        id=None,
        config=None,
        bulk_class=ElasticSearchBulk,
        data_feeder=data_feeder_create_or_index,
    ):
        """
        Description:

        **Parameters**

        app : Application
                        Name of the Application

        pipeline : Pipeline
                        Name of the Pipeline

        connection : Connection
                        Name of the Connection

        id : ID, default= None
                        ID

        config : JSON, default= None
                        Configuration file with additional information.

        bulk_class=ElasticBulk :

        data_feeder=data_feeder_create_or_index :

        |

        """
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)
        self.BulkClass = bulk_class

        self.Index = self.Config.get("index")

        # intex_prefix is obsolete. It is supported currently ensure backward compatibility
        if (
            self.Index == "bspump_"
            and self.Config.get("index_prefix") != "bspump_"
            and len(self.Config.get("index_prefix")) > 0
        ):
            L.warning(
                "The 'index_prefix' has been renamed to 'index', please adjust the configuration."
            )
            self.Index = self.Config.get("index_prefix")

        if data_feeder is None:
            raise RuntimeError("data_feeder must not be None.")

        self.__data_feeder = data_feeder

        app.PubSub.subscribe(
            "ElasticSearchConnection.pause!", self._connection_throttle
        )
        app.PubSub.subscribe(
            "ElasticSearchConnection.unpause!", self._connection_throttle
        )

    def process(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                        Information with timestamp.


        """
        try:
            _id = event.pop("_id", None)
        except TypeError:
            if isinstance(event, dict) is False:
                L.error(
                    "You are trying to pass event of type: {} to ElasticSearchSink, but only dict is supported".format(
                        type(event)
                    )
                )
            raise
        self.Connection.consume(
            context.get("es_index", self.Index),
            self.__data_feeder(event, _id),
            bulk_class=self.BulkClass,
        )

    def _connection_throttle(self, event_name, connection):
        if connection != self.Connection:
            return

        if event_name == "ElasticSearchConnection.pause!":
            self.Pipeline.throttle(self, True)
        elif event_name == "ElasticSearchConnection.unpause!":
            self.Pipeline.throttle(self, False)
        else:
            raise RuntimeError("Unexpected event name '{}'".format(event_name))



from netaddr import IPAddress

import netaddr.core

from bspump.declarative.abc import Expression, evaluate


class IP_PARSE(Expression):
    """
    Parses string, hex number or decimal number to IP address.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Subnet": ["*"],  # TODO: This ...
    }

    Category = "IP"

    # TODO: IP.PARSE should be able to receive expressions as values natively
    def __init__(self, app, *, value=None, arg_value=None):
        super().__init__(app)
        self.Value = value if value is not None else arg_value

    def __call__(self, context, event, *args, **kwargs):
        try:
            ip = IPAddress(evaluate(self.Value, context, event, *args, **kwargs))
        except netaddr.core.AddrFormatError:
            # IP address could not be detected
            return None

        return ip.ipv6().value



import asyncio
import logging

from aioftp import StatusCodeError

from bspump.abc.source import TriggerSource

###

L = logging.getLogger(__name__)

###


class FTPSource(TriggerSource):
    """
    Description:

    |

    """

    ConfigDefaults = {
        "remote_path": "/",
        "mode": "r",  # r = read, p = pop
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application

        pipeline :

        connection :

        id : ID, default = None

        config : JSON, default = None


        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop
        self.Pipeline = pipeline
        self.Queue = asyncio.Queue()
        self.Connection = pipeline.locate_connection(app, connection)
        self.Filename = self.Config.get("filename", None)
        self.RemotePath = self.Config["remote_path"]
        self._conn_future = None
        self.list_future = None

    async def list_files(self):
        """
        Description:

        |

        """

        if self.Queue.qsize() == 0:
            self.list_future = None

        # conenct to the client
        self.client = await self.Connection.connect()

        # if the filename is specified then add only filename to queue else add the
        # full path into queue.

        if self.Filename is not None:
            tmp = self.RemotePath + "/" + self.Filename
            self.Queue.put_nowait(tmp)
        else:
            # if there are directories then don't add them to queue.
            for path, info in await self.client.list(self.RemotePath):
                try:
                    if info["type"] == "dir":
                        continue
                    else:
                        self.Queue.put_nowait(path)
                except asyncio.queues.QueueFull:
                    L.warning("Queue has reached it's limit")

    async def inbound(self):
        """
        Description:

        |

        """
        while True:
            try:
                path = await self.Queue.get()
                async with self.client.download_stream(path) as stream:
                    async for block in stream.iter_by_block():
                        await self.process(block)

                if self.Queue.empty():
                    await self.client.quit()
                    self._conn_future = None
                    break
            except (ConnectionResetError, StatusCodeError) as exception:
                L.exception(exception)

    async def cycle(self):
        """
        Description:

        |

        """
        if self.list_future:
            pass

        else:
            self.list_future = asyncio.ensure_future(
                self.list_files(),
            )

        if self._conn_future:
            pass

        else:
            pass
            self._conn_future = asyncio.ensure_future(
                self.inbound(),
            )
            await self._conn_future



from bspump.declarative.abc import Expression

from ..value.eventexpr import ARG
from ..value.valueexpr import VALUE

import logging

#

L = logging.getLogger(__name__)

#


class CAST(Expression):
    """
    Casts "value" to "type"
    """

    Attributes = {
        "What": ["*"],
        "Default": ["*"],
    }

    CastMap = {
        "bool": (bool, bool.__name__),
        "ui1": (bool, bool.__name__),
        "si1": (bool, bool.__name__),
        "int": (int, int.__name__),
        "si8": (int, "si8"),
        "si16": (int, "si16"),
        "si32": (int, "si32"),
        "si64": (int, "si64"),
        "si128": (int, "si128"),
        "si256": (int, "si256"),
        "ui8": (int, "ui8"),
        "ui16": (int, "ui16"),
        "ui32": (int, "ui32"),
        "ui64": (int, "ui64"),
        "ui128": (int, "ui128"),
        "ui256": (int, "ui256"),
        "float": (float, float.__name__),
        "fp16": (float, float.__name__),
        "fp32": (float, float.__name__),
        "fp64": (float, float.__name__),
        "fp128": (float, float.__name__),
        "str": (str, str.__name__),
        "dict": (dict, dict.__name__),
        "list": (list, list.__name__),
    }

    def __init__(
        self, app, *, arg_what=None, arg_type=None, arg_default=None, value=None
    ):
        super().__init__(app)

        if value is not None:
            # Scalar variant
            # TODO (Dec 2020, AT): This one (scalar variant) is weird, not sure if used anywhere

            self.What = ARG(app=app, value="")
            self.Conversion, self.OutletType = self.CastMap[value]

        else:
            if isinstance(arg_what, Expression):
                self.What = arg_what
            else:
                self.What = VALUE(app, value=arg_what)

            # Detect type cast function
            self.Conversion, self.OutletType = self.CastMap[arg_type]

        if isinstance(arg_default, Expression):
            self.Default = arg_default
        else:
            self.Default = VALUE(app, value=arg_default)

    def __call__(self, context, event, *args, **kwargs):
        try:
            return self.Conversion(self.What(context, event, *args, **kwargs))
        except (ValueError, AttributeError, TypeError):
            if self.Default is None:
                return None
            return self.Default(context, event, *args, **kwargs)

    def get_outlet_type(self):
        return self.OutletType

    def consult_inlet_type(self, key, child):
        return self.OutletType



import aiohttp.web
import copy
import fnmatch
import datetime

from .openmetric import metric_to_openmetric
from ..web.rest import json_response
from ..web.auth import noauth


class MetricWebHandler(object):
    def __init__(self, metrics_svc, webapp):
        self.MetricsService = metrics_svc
        self.App = self.MetricsService.App

        # Add routes
        webapp.router.add_get("/asab/v1/metrics", self.metrics)
        webapp.router.add_get("/asab/v1/watch_metrics", self.watch)
        webapp.router.add_get("/asab/v1/metrics.json", self.metrics_json)

    @noauth
    async def metrics_json(self, request):
        """
        Get metrics in a JSON.
        ---
        tags: ['asab.metrics']
        """
        metrics_to_send = copy.deepcopy(self.MetricsService.Storage.Metrics)
        return json_response(request, metrics_to_send)

    @noauth
    async def metrics(self, request):
        """
        Produce the OpenMetrics output.
        ---
        tags: ['asab.metrics']
        """
        lines = []

        for data in self.MetricsService.Storage.Metrics:
            line = metric_to_openmetric(data)
            if line is not None:
                lines.append(line)

        if lines:
            lines.append("# EOF\n")

        text = "\n".join(lines)

        return aiohttp.web.Response(
            text=text,
            content_type="text/plain",
            charset="utf-8",
        )

    @noauth
    async def watch(self, request):
        """
        Endpoint to list ASAB metrics in the command line.

        Example commands:
        * watch curl localhost:8080/asab/v1/watch_metrics -> all metrics

        Strict filters (inclusion and exclusion):
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=web_requests -> web_requests metric only
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=-web_requests -> all metrics w/o web_requests metric

        Includes only:
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=web* -> metrics that start with "web"
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=*web -> metrics that end with "web"
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=*web* -> metrics that contain "web"

        Excludes:
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=-web* -> metrics that start with "web"
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=-*web -> metrics that end with "web"
        * watch curl localhost:8080/asab/v1/watch_metrics?filter=-*web* -> metrics that contain "web"

        ---
        tags: ['asab.metrics']
        """

        filter = request.query.get("filter")
        tags = request.query.get("tags")
        tags = True if tags is not None and tags.lower() == "true" else False
        text = watch_table(self.MetricsService.Storage.Metrics, filter, tags)

        return aiohttp.web.Response(
            text=text,
            content_type="text/plain",
            charset="utf-8",
        )


def watch_table(metric_records: list, filter, tags):
    lines = []
    m_name_len = max([len(i["name"]) for i in metric_records])

    v_name_len = 0
    for i in metric_records:
        if len(i["fieldset"]) == 0:
            continue
        if len(i["fieldset"][0].get("values", {})) == 0:
            continue
        metric_name_len = (
            max(
                [
                    len(str(value_name))
                    for value_name in i["fieldset"][0].get("values").keys()
                ]
            )
            + 10
        )
        if metric_name_len > v_name_len:
            v_name_len = metric_name_len

    timestamp_len = 30
    t_name_len = max(
        [len(str(field["tags"])) for i in metric_records for field in i["fieldset"]]
    )

    if tags:
        separator = "-" * (
            m_name_len + v_name_len + t_name_len + timestamp_len + 30 + 2
        )
    else:
        separator = "-" * (m_name_len + v_name_len + timestamp_len + 30 + 2)

    lines.append(separator)
    lines.append(
        build_line(
            "Metric name",
            "Value name",
            "Value",
            "Timestamp",
            m_name_len,
            v_name_len,
            tags,
            timestamp_len,
            t_string="Tags",
            t_name_len=t_name_len,
        )
    )
    lines.append(separator)

    for metric_record in metric_records:
        for field in metric_record["fieldset"]:
            if field.get("values") is None:
                continue
            name = metric_record.get("name")
            timestamp = field.get("measured_at")
            timestamp = datetime.datetime.fromtimestamp(timestamp)

            if filter is not None:
                if filter.startswith("-"):
                    if fnmatch.fnmatch(name, filter[1:]):
                        continue
                else:
                    if not fnmatch.fnmatch(name, filter):
                        continue

            if metric_record.get("type") in ["Histogram", "HistogramWithDynamicTags"]:
                for upperboud, values in field.get("values").get("buckets").items():
                    for v_name, value in values.items():
                        lines.append(
                            build_line(
                                str(name),
                                str(v_name),
                                str(value),
                                str(timestamp),
                                m_name_len,
                                v_name_len,
                                tags,
                                timestamp_len,
                                str(upperboud),
                                t_string=str(field["tags"]),
                                t_name_len=t_name_len,
                            )
                        )

                lines.append(
                    build_line(
                        str(name),
                        "Sum",
                        field.get("values").get("sum"),
                        str(timestamp),
                        m_name_len,
                        v_name_len,
                        tags,
                        timestamp_len,
                        t_string=str(field["tags"]),
                        t_name_len=t_name_len,
                    )
                )
                lines.append(
                    build_line(
                        str(name),
                        "Count",
                        field.get("values").get("count"),
                        str(timestamp),
                        m_name_len,
                        v_name_len,
                        tags,
                        timestamp_len,
                        t_string=str(field["tags"]),
                        t_name_len=t_name_len,
                    )
                )

            else:
                for key, value in field.get("values").items():
                    lines.append(
                        build_line(
                            str(name),
                            str(key),
                            str(value),
                            str(timestamp),
                            m_name_len,
                            v_name_len,
                            tags,
                            timestamp_len,
                            t_string=str(field["tags"]),
                            t_name_len=t_name_len,
                        )
                    )

    text = "\n".join(lines)
    return text


def build_line(
    name,
    value_name,
    value,
    timestamp,
    m_name_len,
    v_name_len,
    tags,
    timestamp_len,
    upperbound=None,
    t_string=None,
    t_name_len=None,
):
    if upperbound is not None:
        if tags:
            line = "{:<{m_name_len}} | {:<{t_name_len}} | {:<{v_name_len}} | {:<7} | {:<25} | {:<{timestamp_len}}".format(
                name,
                t_string,
                value_name,
                upperbound,
                value,
                timestamp,
                v_name_len=v_name_len - 10,
                m_name_len=m_name_len,
                t_name_len=t_name_len,
                timestamp_len=timestamp_len,
            )
        else:
            line = "{:<{m_name_len}} | {:<{v_name_len}} | {:<7} | {:<25} | {:<{timestamp_len}}".format(
                name,
                value_name,
                upperbound,
                value,
                timestamp,
                v_name_len=v_name_len - 10,
                m_name_len=m_name_len,
                timestamp_len=timestamp_len,
            )
    else:
        if tags:
            line = "{:<{m_name_len}} | {:<{t_name_len}} | {:<{v_name_len}} | {:<25} | {:<{timestamp_len}}".format(
                name,
                t_string,
                value_name,
                value,
                timestamp,
                v_name_len=v_name_len,
                m_name_len=m_name_len,
                t_name_len=t_name_len,
                timestamp_len=timestamp_len,
            )
        else:
            line = "{:<{m_name_len}} | {:<{v_name_len}} | {:<25} | {:<{timestamp_len}}".format(
                name,
                value_name,
                value,
                timestamp,
                v_name_len=v_name_len,
                m_name_len=m_name_len,
                timestamp_len=timestamp_len,
            )

    return line



import os
import logging
import hashlib
import random
import tarfile
import asyncio
import shutil
import tempfile
import urllib.parse

import aiohttp

from .filesystem import FileSystemLibraryProvider
from ..dirsync import synchronize_dirs

#

L = logging.getLogger(__name__)

#


class LibsRegLibraryProvider(FileSystemLibraryProvider):
    """
    Read-only provider to read from remote "library repository".

    It provides an option to specify more servers for more reliable content delivery.

    Example of the configuration:

    ```ini
    [library]
    providers=
    ...
    libsreg+https://libsreg1.example.com,libsreg2.example.com/my-library
    ...
    ```

    """

    def __init__(self, library, path, layer):
        url = urllib.parse.urlparse(path)
        assert url.scheme.startswith("libsreg+")

        version = url.fragment
        if version == "":
            version = "production"

        archname = url.path[1:]

        self.URLs = [
            "{scheme}://{netloc}/{archname}/{archname}-{version}.tar.xz".format(
                scheme=url.scheme[8:],
                netloc=netloc,
                archname=archname,
                version=version,
            )
            for netloc in url.netloc.split(",")
        ]
        assert len(self.URLs) > 0

        tempdir = tempfile.gettempdir()
        self.RootPath = os.path.join(
            tempdir,
            "asab.library.libsreg",
        )

        self.RepoPath = os.path.join(
            self.RootPath, hashlib.sha256(self.URLs[0].encode("utf-8")).hexdigest()
        )

        os.makedirs(os.path.join(self.RepoPath), exist_ok=True)

        super().__init__(library, self.RepoPath, layer, set_ready=False)

        self.PullLock = asyncio.Lock()

        # TODO: Subscribption to changes in the library
        self.SubscribedPaths = set()

        self.App.TaskService.schedule(self._periodic_pull(None))
        self.App.PubSub.subscribe("Application.tick/60!", self._periodic_pull)

    async def _periodic_pull(self, event_name):
        """
        Changes in remote repository are being pulled every minute.
        `PullLock` ensures that only if previous "pull" has finished, new one can start.
        """

        if self.PullLock.locked():
            return

        async with self.PullLock:
            headers = {}

            # Check for existing E-Tag
            etag_fname = os.path.join(self.RootPath, "etag")
            try:
                with open(etag_fname, "r") as f:
                    etag = f.read().strip()
                    headers["If-None-Match"] = etag
            except FileNotFoundError:
                pass

            url = random.choice(self.URLs)

            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(url, headers=headers) as response:
                        if (
                            response.status == 200
                        ):  # The request indicates a new version that we don't have yet
                            etag_incoming = response.headers.get("ETag")

                            # Download new version
                            newtarfname = os.path.join(self.RootPath, "new.tar.xz")
                            with open(newtarfname, "wb") as ftmp:
                                while True:
                                    chunk = await response.content.read(16 * 1024)
                                    if not chunk:
                                        break
                                    ftmp.write(chunk)

                            # Extract the contents to the temporary directory
                            temp_extract_dir = os.path.join(self.RootPath, "new")

                            # Remove temp_extract_dir if it exists (from the last, failed run)
                            if os.path.exists(temp_extract_dir):
                                shutil.rmtree(temp_extract_dir)

                            # Extract the archive into the temp_extract_dir
                            with tarfile.open(newtarfname, mode="r:xz") as tar:
                                tar.extractall(temp_extract_dir)

                            # Synchronize the temp_extract_dir into the library
                            synchronize_dirs(self.RepoPath, temp_extract_dir)
                            if not self.IsReady:
                                await self._set_ready()

                            if etag_incoming is not None:
                                with open(etag_fname, "w") as f:
                                    f.write(etag_incoming)

                            # Remove temp_extract_dir
                            if os.path.exists(temp_extract_dir):
                                shutil.rmtree(temp_extract_dir)

                            # Remove newtarfname
                            if os.path.exists(newtarfname):
                                os.remove(newtarfname)

                        elif response.status == 304:
                            # The repository has not changed ...
                            if not self.IsReady:
                                await self._set_ready()

                        else:
                            L.error(
                                "Failed to download the library.",
                                struct_data={"url": url, "status": response.status},
                            )

            except aiohttp.ClientError as e:
                L.error(
                    "Failed to download the library.",
                    struct_data={"url": url, "exception": e},
                )

            except asyncio.TimeoutError as e:
                L.error(
                    "Failed to download the library.",
                    struct_data={"url": url, "exception": e},
                )

    async def subscribe(self, path):
        self.SubscribedPaths.add(path)



import aiohttp.web


def authz_middleware_factory(app, svc):
    """
    OBSOLETE

    Ensures that AuthzService is part of the request.
    :param app: application object
    :param svc: AuthzService
    :return: handler(request)
    """

    @aiohttp.web.middleware
    async def authz_middleware(request, handler):
        request.AuthzService = svc
        return await handler(request)

    return authz_middleware



import os
import uuid
import json
import datetime
import logging

from .. import Service, Config
from ..utils import running_in_container
from .web_handler import APIWebHandler
from .log import WebApiLoggingHandler
from .doc import DocWebHandler
from .discovery import DiscoveryService

##

L = logging.getLogger(__name__)


##


class ApiService(Service):
    def __init__(self, app, service_name="asab.ApiService"):
        super().__init__(app, service_name)

        self.WebContainer = None
        self.ZkContainer = None
        self.MetricWebHandler = None

        self.AttentionRequired = {}  # dict of errors found.

        # Manifest
        path = Config.get("general", "manifest")
        if path == "":
            if os.path.isfile("/app/MANIFEST.json"):
                path = "/app/MANIFEST.json"
            elif os.path.isfile("/MANIFEST.json"):
                path = "/MANIFEST.json"
            elif os.path.isfile("MANIFEST.json"):
                path = "MANIFEST.json"

        if len(path) != 0:
            with open(path) as f:
                self.Manifest = json.load(f)

        else:
            self.Manifest = None

        # Change log
        path = Config.get("general", "changelog")
        if path == "":
            if os.path.isfile("/app/CHANGELOG.md"):
                path = "/app/CHANGELOG.md"
            elif os.path.isfile("/CHANGELOG.md"):
                path = "/CHANGELOG.md"
            elif os.path.isfile("CHANGELOG.md"):
                path = "CHANGELOG.md"

        if os.path.isfile(path):
            self.ChangeLog = path
        else:
            self.ChangeLog = None

        # Service Discovery
        self.DiscoveryService = None

    def attention_required(self, att: dict, att_id=None):
        if att_id is None:
            # add new attention id to list
            att_id = uuid.uuid4().hex
        self.AttentionRequired[att_id] = att

        # if creation time for att_id is not present then add
        if "_c" not in att:
            att["_c"] = (
                datetime.datetime.utcnow().isoformat() + "Z"
            )  # This is OK, no tzinfo needed

        # add to microservice json/dict section attention_required
        self._do_zookeeper_adv_data()
        return att_id

    def remove_attention(self, att_id):
        try:
            # find the attention id value and remove it.
            for error_key, error_value in self.AttentionRequired.items():
                if error_key == att_id:
                    del self.AttentionRequired[att_id]
                    break
        except KeyError:
            L.warning("Key None does not exist.")
            raise Exception("Key None does not exist.")

        self._do_zookeeper_adv_data()

    def initialize_web(self, webcontainer=None):
        """
        Example:

        websvc = self.App.get_service("asab.WebService")
        webcontainer = asab.web.WebContainer(websvc, 'web')
        ApiService.initialize_web(webcontainer)

        Initialize into a default web container:
        ApiService.initialize_web()
        """
        assert self.WebContainer is None

        if webcontainer is None:
            websvc = self.App.get_service("asab.WebService")
            webcontainer = websvc.WebContainer

        self.WebContainer = webcontainer

        # TODO: Logging level configurable via config file
        self.APILogHandler = WebApiLoggingHandler(self.App, level=logging.NOTSET)
        self.format = logging.Formatter(
            "%%(asctime)s %%(levelname)s %%(name)s %%(struct_data)s%%(message)s"
        )
        self.APILogHandler.setFormatter(self.format)
        self.Logging = logging.getLogger()
        self.Logging.addHandler(self.APILogHandler)

        self.WebHandler = APIWebHandler(
            self, self.WebContainer.WebApp, self.APILogHandler
        )

        self.DocWebHandler = DocWebHandler(self, self.App, self.WebContainer)

        # If asab.MetricsService is available, initialize its web handler
        metrics_svc = self.App.get_service("asab.MetricsService")
        if metrics_svc is not None:
            from ..metrics.web_handler import MetricWebHandler

            self.MetricWebHandler = MetricWebHandler(
                metrics_svc, self.WebContainer.WebApp
            )

        self.App.PubSub.subscribe("WebContainer.started!", self._on_webcontainer_start)

    def initialize_zookeeper(self, zoocontainer=None):
        """
        Example:

        Initialize into a default zookeeper container:

        ```
        ApiService.initialize_zookeeper()
        ```


        Intialize with a custom Zookeeper container:

        ```
        zksvc = self.App.get_service("asab.ZooKeeperService")
        zoocontainer = ZooKeeperContainer(zksvc, "mycontainer")
        ApiService.initialize_zookeeper(zoocontainer)
        ```

        """
        assert self.ZkContainer is None

        if zoocontainer is None:
            zksvc = self.App.get_service("asab.ZooKeeperService")
            from ..zookeeper import ZooKeeperContainer

            zoocontainer = ZooKeeperContainer(zksvc)

        # get zookeeper-service
        self.ZkContainer = zoocontainer

        # initialize service discovery
        self.DiscoveryService = DiscoveryService(self.App, self.ZkContainer)

        self.App.PubSub.subscribe(
            "ZooKeeperContainer.state/CONNECTED!", self._on_zkcontainer_start
        )
        self._do_zookeeper_adv_data()

    def _do_zookeeper_adv_data(self):
        if self.ZkContainer is None:
            return

        if not self.ZkContainer.is_connected():
            return

        adv_data = {
            "host": self.App.HostName,
            "appclass": self.App.__class__.__name__,
            "launch_time": datetime.datetime.utcfromtimestamp(
                self.App.LaunchTime
            ).isoformat()
            + "Z",
            "process_id": os.getpid(),
        }

        # A unique identifier of a microservice; added as an environment variable
        instance_id = os.getenv("INSTANCE_ID", None)
        if instance_id is not None:
            adv_data["instance_id"] = instance_id

        # A identified of the host machine (node); added if available at environment variables
        node_id = os.getenv("NODE_ID", None)
        if node_id is not None:
            adv_data["node_id"] = node_id

        service_id = os.getenv("SERVICE_ID", None)
        if service_id is not None:
            adv_data["service_id"] = service_id

        if running_in_container():
            adv_data["containerized"] = True

        if self.Manifest is not None:
            adv_data.update(self.Manifest)

        if len(self.AttentionRequired) > 0:
            # add attention required status
            adv_data.update({"attention_required": self.AttentionRequired})

        if self.WebContainer is not None:
            adv_data["web"] = self.WebContainer.Addresses

        self.ZkContainer.advertise(
            data=adv_data,
            path="/run/{}.".format(self.App.__class__.__name__),
        )

    def _on_zkcontainer_start(self, message_type, container):
        if container == self.ZkContainer:
            self._do_zookeeper_adv_data()

    def _on_webcontainer_start(self, message_type, container):
        if container == self.WebContainer:
            self._do_zookeeper_adv_data()



import collections
import os
import shelve


class PersistentDict(dict):
    """
    The persistent dictionary works as the regular Python dictionary but the content of the dictionary is stored in the file.
    You cat think of a `PersistentDict` as a simple [key-value store](https://en.wikipedia.org/wiki/Key-value_database).
    It is not optimized for a frequent access. This class provides common `dict` interface.

    Example:
            ```python
            class MyApplication(asab.Application):
                    async def main(self):
                            pdict = asab.PersistentDict('./pdict.bin')
                            pdict.load()
                            counter = pdict['counter'] = pdict.setdefault('counter', 0) + 1
                            print("Executed for {} times".format(counter))
                            pdict.store()
                            self.stop()
            ```

    !!! warning
            You must explicitly `load()` and `store()` content of the dictionary!

    !!! warning
            You can only store objects in the persistent dictionary that are serializable.
    """

    def __init__(self, path: str):
        """Initialize persistent dictionary.

        Args:
                path (str): Path for the dictionary file.
        """
        super().__init__()
        # Create directory, if needed
        dirname = os.path.dirname(path)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        self._path = path

    def __delitem__(self, key):
        super().__delitem__(key)
        with shelve.open(self._path) as d:
            del d[key]

    def load(self) -> None:
        """
        Load content of file as dictionary.
        """
        with shelve.open(self._path) as d:
            for key, value in d.items():
                self[key] = value

    def store(self) -> None:
        """
        Store content of persistent dictionary to a file.
        """

        with shelve.open(self._path) as d:
            for key, value in self.items():
                d[key] = value

    def update(self, other={}, **kwds) -> None:
        """
        Update persistent dictionary from mapping or iterable.

        Examples:
                ```python
                        >>> pdict.update({'foo': 'bar', 'moo': 'buzz'})
                        >>> pdict.update(foo='bar', moo='buzz')
                        >>> pdict.update([('foo','bar'),('moo','buzz')])
                ```

        Args:
                other: Dictionary or iterable of 2-tuples of the form (key, value) to be updated.
        """

        if isinstance(other, collections.abc.Mapping):
            for key in other:
                self[key] = other[key]
        elif hasattr(other, "keys"):
            for key in other.keys():
                self[key] = other[key]
        else:
            for key, value in other:
                self[key] = value

        for key, value in kwds.items():
            self[key] = value



import fastavro
from ..avro import loader
from ..file.fileabcsource import FileABCSource
import logging


###

L = logging.getLogger(__name__)

###


class AvroSource(FileABCSource):
    ConfigDefaults = {
        "path": "./*.avro",
        "post": "noop",  # one of 'delete', 'noop' and 'move'
        "schema_file": "",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Schema = loader.load_avro_schema(self.Config)

    async def read(self, filename, f):
        while True:
            if self.Schema is None:
                L.warning("Schema file is not provided.")
            else:
                L.warning("Schema file is used.")

            avro_reader = fastavro.reader(f, self.Schema)
            for record in avro_reader:
                await self.process(record, {"filename": filename})



from .client.source import HTTPClientLineSource
from .client.source import HTTPClientSource
from .client.source import HTTPClientTextSource
from .client.wssink import HTTPClientWebSocketSink
from .web.sink import WebServiceSink
from .web.source import WebServiceSource
from .web.wssource import WebSocketSource
from .lookupprovider import HTTPBatchLookupProvider

__all__ = (
    "HTTPClientSource",
    "HTTPClientTextSource",
    "HTTPClientLineSource",
    "HTTPClientWebSocketSink",
    "WebServiceSource",
    "WebServiceSink",
    "WebSocketSource",
    "HTTPBatchLookupProvider",
)



import logging
import aiohttp
import urllib
import http.client

import bspump.asab as asab

#

L = logging.getLogger(__name__)


#


class InfluxDBTarget(asab.Configurable):
    """
    InfluxDB 2.0 API parameters:

    - url - [required] url string of your influxDB
    - bucket - [required] the destination bucket for writes
    - org - [required] the parameter value specifies the destination organization for writes
    - orgid - [optional] the parameter value specifies the ID of the destination organization for writes
    - token - [required] API token to authenticate to the InfluxDB

    NOTE: If both orgID and org are specified, org takes precedence

    Example:
    ```ini
    [asab:metrics:influxdb]
    url=http://localhost:8086
    bucket=test
    org=test
    orgid=test
    token=your_token
    ```

    InfluxDB <1.8 API parameters:
    - url - [required] url string of your influxDB
    - username - [required] name of influxDB user
    - password - [required] password of influxDB user

    Example:

    ```ini
    [asab:metrics:influxdb]
    url=http://localhost:8086
    username=test
    password=testtest
    db=test
    ```
    """

    ConfigDefaults = {
        "url": "http://localhost:8086/",
        "db": "mydb",
        "username": "",
        "password": "",
        "proactor": True,  # Use ProactorService to send metrics on thread
    }

    def __init__(self, svc, config_section_name, config=None):
        super().__init__(config_section_name=config_section_name, config=config)
        self.Headers = {}
        self.BaseURL = self.Config.get("url").rstrip("/")
        self.WriteRequest = "/write?db={}".format(self.Config.get("db"))

        username = self.Config.get("username")
        if username is not None and len(username) > 0:
            self.WriteRequest += "&u={}".format(urllib.parse.quote(username, safe=""))

        password = self.Config.get("password")
        if password is not None and len(password) > 0:
            self.WriteRequest += "&p={}".format(urllib.parse.quote(password, safe=""))

        # If org is specified we are buildig write request for InfluxDB 2.0 API
        org = self.Config.get("org")
        if org is not None:
            self.WriteRequest = "/api/v2/write?org={}".format(org)

        bucket = self.Config.get("bucket")
        if bucket is not None:
            self.WriteRequest += "&bucket={}".format(bucket)

        orgid = self.Config.get("orgid")
        if orgid is not None:
            self.WriteRequest += "&orgID={}".format(orgid)

        token = self.Config.get("token")
        if token is not None:
            self.Headers = {"Authorization": "Token {}".format(token)}

        self.WriteURL = "{}{}".format(self.BaseURL, self.WriteRequest)

        # Proactor service is used for alternative delivery of the metrics into the InfluxDB
        # It is handly when a main loop can become very busy
        if self.Config.getboolean("proactor"):
            try:
                from ..proactor import Module

                svc.App.add_module(Module)
                self.ProactorService = svc.App.get_service("asab.ProactorService")
            except KeyError:
                self.ProactorService = None
        else:
            self.ProactorService = None

    async def process(self, m_tree, now):
        rb = influxdb_format(m_tree, now)

        if self.ProactorService is not None:
            self.ProactorService.schedule(self._worker_upload, m_tree, rb)

        else:
            try:
                async with aiohttp.ClientSession(headers=self.Headers) as session:
                    async with session.post(self.WriteURL, data=rb) as resp:
                        response = await resp.text()
                        if resp.status != 204:
                            L.warning(
                                "Error when sending metrics to Influx: {}\n{}".format(
                                    resp.status, response
                                ),
                                struct_data={"url": self.BaseURL},
                            )
            except aiohttp.client_exceptions.ClientConnectorError:
                L.error(
                    "Failed to connect to InfluxDB.", struct_data={"url": self.BaseURL}
                )
            except Exception as err:
                print(
                    "Failed to send metrics to InfluxDB: {}".format(err),
                    struct_data={"url": self.BaseURL},
                )

    def _worker_upload(self, m_tree, rb):
        if self.BaseURL.startswith("https://"):
            conn = http.client.HTTPSConnection(self.BaseURL.replace("https://", ""))
        else:
            conn = http.client.HTTPConnection(self.BaseURL.replace("http://", ""))

        try:
            conn.request("POST", self.WriteRequest, rb, self.Headers)
        except ConnectionError:
            L.error("Failed to connect to InfluxDB.", struct_data={"url": self.BaseURL})
            return
        except Exception as err:
            L.exception(
                "Failed to send metrics to InfluxDB: {}".format(err),
                struct_data={"url": self.BaseURL},
            )

        response = conn.getresponse()
        if response.status != 204:
            L.warning(
                "Sending metrics to InfluxDB failed.",
                struct_data={
                    "url": self.BaseURL,
                    "response.status": response.status,
                    "response": response.read().decode("utf-8"),
                },
            )


def get_field(fk, fv):
    if isinstance(fv, bool):
        field = "{}={}".format(fk, "t" if fv else "f")
    elif isinstance(fv, int):
        field = "{}={}i".format(fk, fv)
    elif isinstance(fv, float):
        field = "{}={}".format(fk, fv)
    elif isinstance(fv, str):
        # Escapes the Field Values and Field Keys if the value is a string
        field = '{}="{}"'.format(
            fk.replace(" ", r"\ ").replace(",", r"\,").replace("=", r"\="),
            fv.replace("\\", "\\\\").replace('"', '\\"'),
        )
    else:
        raise RuntimeError(
            "Unknown/invalid type of the metrics field: {} {}".format(type(fv), fk)
        )

    return field


def combine_tags_and_field(tags, values, timestamp):
    # First escape tags and values
    tags = escape_tags(tags)
    values = escape_values(values)
    # Then combine the tags and then values
    # remove "help" and "unit" tags -> utilized in openmetric target
    tags_string = ",".join(
        [
            "{}={}".format(tk, tv)
            for tk, tv in tags.items()
            if tk not in ("help", "unit")
        ]
    )
    field_set = ",".join(
        [get_field(value_name, value) for value_name, value in values.items()]
    )
    return tags_string + " " + field_set + " " + str(int(timestamp * 1e9))


def build_metric_line(tags, values, timestamp, upperbound=None):
    if upperbound is not None:
        tags["le"] = upperbound
    return combine_tags_and_field(tags, values, timestamp)


def get_timestamp(field, now):
    if "measured_at" in field:
        timestamp = field["measured_at"]
    else:
        timestamp = now
    return timestamp


def metric_to_influxdb(metric_record, now):
    name = escape_name(metric_record.get("name"))
    fieldset = metric_record.get("fieldset")
    metric_type = metric_record.get("type")
    values_lines = []

    if metric_type in ["Histogram", "HistogramWithDynamicTags"]:
        for field in fieldset:
            # SKIP empty fields
            if all(
                [bucket == {} for bucket in field.get("values").get("buckets").values()]
            ):
                continue
            timestamp = get_timestamp(field, now)
            for upperbound, bucket in field.get("values").get("buckets").items():
                upperbound = str(upperbound)
                if bucket == {}:
                    continue
                values_lines.append(
                    build_metric_line(
                        field.get("tags").copy(), bucket, timestamp, upperbound
                    )
                )
            values_lines.append(
                build_metric_line(
                    field.get("tags").copy(),
                    {"sum": field.get("values").get("sum")},
                    timestamp,
                )
            )
            values_lines.append(
                build_metric_line(
                    field.get("tags").copy(),
                    {"count": field.get("values").get("count")},
                    timestamp,
                )
            )

    else:
        for field in fieldset:
            # SKIP empty fields
            if not field.get("values") or field.get("values") == {}:
                continue
            timestamp = get_timestamp(field, now)
            values_lines.append(
                build_metric_line(field.get("tags"), (field.get("values")), timestamp)
            )

    return ["{},{}\n".format(name, line) for line in values_lines]


def escape_name(name: str):
    return name.replace(" ", "\\ ").replace(",", "\\,")


def escape_tags(tags: dict):
    """
    Escapes special characters in inputted tags to comply with InfluxDB's rules

    https://docs.influxdata.com/influxdb/v2.3/reference/syntax/line-protocol/#special-characters
    """
    clean: dict = {}
    for k, v in tags.items():
        if v is None:
            v = "unknown"
        clean[k.replace(" ", "\\ ").replace(",", "\\,").replace("=", "\\=")] = (
            v.replace(" ", "\\ ").replace(",", "\\,").replace("=", "\\=")
        )
    return clean


def escape_values(values: dict):
    """
    Escapes special characters in inputted values to comply with InfluxDB's rules

    https://docs.influxdata.com/influxdb/v2.3/reference/syntax/line-protocol/#special-characters
    """
    clean: dict = {}
    for k, v in values.items():
        # Escapes the Field Keys
        clean[k.replace(" ", r"\ ").replace(",", r"\,").replace("=", r"\=")] = v
    return clean


def influxdb_format(m_tree, now):
    rb = []
    for metric_record in m_tree:
        influx_records = metric_to_influxdb(metric_record, now)
        rb.extend(influx_records)
    return "".join(rb)



import unittest
import collections
import time

import bspump
import bspump.matrix
import bspump.unittest


class TestNamedMatrix(bspump.unittest.TestCase):
    def test_matrix_zeros(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App, dtype="int_")

        matrix.zeros()
        self.assertEqual(matrix.Array.shape, (1,))
        self.assertEqual(len(matrix.Index.N2IMap), 0)
        self.assertEqual(len(matrix.Index.I2NMap), 0)

    def test_matrix_flush(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App)
        n = 100
        indexes = []
        for i in range(n):
            index = matrix.add_row(str(i))
            indexes.append(index)
            matrix.Array[index] = i

        check_array = matrix.Array[40:100]

        for i in range(0, 40):
            matrix.close_row(str(i))

        matrix.flush()
        self.assertEqual(matrix.Array.shape, (n - 40,))
        self.assertEqual(len(matrix.Index.N2IMap), n - 40)
        self.assertEqual(len(matrix.Index.I2NMap), n - 40)
        self.assertEqual(len(check_array), len(matrix.Array))
        for i in range(len(check_array)):
            self.assertEqual(check_array[i], matrix.Array[i])

    def test_matrix_add_row(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App)
        n = 100
        n2i = collections.OrderedDict()
        i2n = collections.OrderedDict()
        for i in range(n):
            name = "id_" + str(i)
            index = matrix.add_row(name)
            n2i[name] = index
            i2n[index] = name

        self.assertEqual(n2i, matrix.Index.N2IMap)
        self.assertEqual(i2n, matrix.Index.I2NMap)

    def test_matrix_close_row(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App)
        n = 100
        for i in range(n):
            index = matrix.add_row(str(i))

        for i in range(0, 5):
            matrix.close_row(str(i))
            self.assertNotIn(i, matrix.Index.I2NMap)

        self.assertEqual(len(matrix.Index.I2NMap), len(matrix.Index.N2IMap))

    def test_matrix_get_row_index(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App)
        n = 100
        for i in range(n):
            name = "id_" + str(i)
            index = matrix.add_row(name)
            index_obtained = matrix.get_row_index(name)
            self.assertEqual(i, index_obtained)

    def test_matrix_get_row_name(self):
        matrix = bspump.matrix.NamedMatrix(app=self.App)
        n = 100
        for i in range(n):
            name = "id_" + str(i)
            index = matrix.add_row(name)
            name_obtained = matrix.get_row_name(i)
            self.assertEqual(name, name_obtained)



import os
import urllib.parse
import configparser
import typing


def convert_to_seconds(value: str) -> float:
    """
    Parse time duration string (e.g. "3h", "20m" or "1y") and convert it into seconds.

    Args:
            value: Time duration string.

    Returns:
            float: Number of seconds.

    Raises:
            ValueError: If the string is not in a valid format.
    """
    if isinstance(value, (int, float)):
        return float(value)

    value = value.replace(" ", "")
    try:
        # Second condition in each IF is for backward compatibility
        if value.endswith("ms"):
            value = float(value[:-2]) / 1000.0
        elif value.endswith("y") or value.endswith("Y"):
            value = float(value[:-1]) * 86400 * 365
        elif value.endswith("M"):
            value = float(value[:-1]) * 86400 * 31
        elif value.endswith("w") or value.endswith("W"):
            value = float(value[:-1]) * 86400 * 7
        elif value.endswith("d") or value.endswith("D"):
            value = float(value[:-1]) * 86400
        elif value.endswith("h"):
            value = float(value[:-1]) * 3600
        elif value.endswith("m"):
            value = float(value[:-1]) * 60
        elif value.endswith("s"):
            value = float(value[:-1])
        else:
            value = float(value)
    except ValueError as e:
        raise ValueError("'{}' is not a valid time specification: {}.".format(value, e))

    return value


def convert_to_bytes(size: str) -> int:
    """
    Convert a size string to bytes. The size string should be a number
    optionally followed by a unit (B, kB, MB, GB, or TB), e.g., "10MB".

    Examples:
            Configuration:
            ```ini
            [general]
            rotate_size=30G
            ```
            Usage:
            ```python
            self.RotateAtSize = asab.utils.convert_to_bytes(asab.Config.get('general', 'rotate_size'))
            ```

    Args:
            size: Size string.

    Returns:
            Size in bytes.

    Raises:
            ValueError: If the size string does not have the correct format.
    """
    units = {
        "B": 1,
        "kB": 10**3,
        "MB": 10**6,
        "GB": 10**9,
        "TB": 10**12,
        # These are typical shortcuts that users take, we support them as well
        "k": 10**3,
        "K": 10**3,
        "M": 10**6,
        "G": 10**9,
        "T": 10**12,
    }
    size = size.strip()  # remove leading and trailing whitespace

    if size.isdigit():
        # size is just a number, so it's already in bytes
        return int(size)

    # size has a unit, find where the number part ends
    for i, char in enumerate(size):
        if not char.isdigit() and char != ".":
            break
    else:
        # no unit found
        raise ValueError("Invalid size string: {}".format(size))

    number = size[:i]
    unit = size[i:].strip()

    if unit not in units:
        raise ValueError("Invalid unit: {}".format(unit))

    return int(float(number) * units[unit])


def string_to_boolean(value: str) -> bool:
    """
    Convert common boolean string values (e.g. 'yes' or 'no') into boolean.

    - `True`: `1`, `'yes'`, `'true'`, `'on'`
    - `False`: `0`, `'no'`, `'false'`, `'off'`

    Args:
            value: A value to be parsed.

    Returns:
            Value converted to bool.
    """
    if isinstance(value, bool):
        return value
    if value.lower() not in configparser.ConfigParser.BOOLEAN_STATES:
        raise ValueError("Not a boolean: {}".format(value))
    return configparser.ConfigParser.BOOLEAN_STATES[value.lower()]


def validate_url(
    input_url: str, scheme: typing.Union[str, typing.Tuple[str], None]
) -> str:
    """Parse URL, remove leading and trailing whitespaces and a trailing slash.
    If `scheme` is specified, check if it matches the `input_url` scheme.

    Args:
            input_url (str): URL to be parsed and validated.
            scheme (str | tuple[str] | None): Requested URL schema.

    Raises:
            ValueError: If `scheme` is specified and is invalid.

    Returns:
            str: Parsed and validated URL.
    """
    # Remove leading and trailing whitespaces before parsing
    url = urllib.parse.urlparse(input_url.strip())

    if url.path.endswith("/"):
        url = url._replace(path=url.path[:-1])

    if scheme is None:  # Scheme doesn't get checked
        return url.geturl()
    elif isinstance(scheme, tuple):  # Supports tuple
        if url.scheme in scheme:
            return url.geturl()
    elif scheme == url.scheme:
        return url.geturl()
    else:
        if url.scheme:
            raise ValueError(
                "'{}' has an invalid scheme: '{}'".format(url.geturl(), url.scheme)
            )
        elif not url.scheme:
            raise ValueError("'{}' does not have a scheme".format(url.geturl()))
        else:
            raise ValueError("'{}' has an invalid scheme".format(url.geturl()))
    return url.geturl()


def running_in_container() -> bool:
    """
    Check if the application is running in Docker or Podman container.

    Returns:
            bool: `True` if the application is running in a container.
    """

    # The process ID is 1 only in the docker/podman container
    if os.getpid() == 1:
        return True

    # This works for older versions of Ubuntu with cgroups v1 and Docker
    if os.path.exists("/.dockerenv") and os.path.isfile("/proc/self/cgroup"):
        with open("/proc/self/cgroup", "r") as f:
            if any("docker" in line for line in f.readlines()):
                return True

    # Since Ubuntu 22.04 linux kernel uses cgroups v2 which do not operate with /proc/self/cgroup file
    # Works only for "overlay" filesystem.
    if os.path.isfile("/proc/self/mountinfo"):
        with open("/proc/self/mountinfo", "r") as f:
            for line in f.readlines():
                # Seek for a root filesystem
                if " / / " not in line:
                    continue

                # Is the root filesystem overlay?
                if " overlay " not in line:
                    continue

                return True

    return False



from netaddr import IPNetwork, IPAddress

import netaddr.core

from bspump.declarative.abc import Expression
from ..value.valueexpr import VALUE


class IP_INSUBNET(Expression):
    """
    Checks if expression is of given subnet.
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Subnet": ["*"],  # TODO: This ...
    }

    Category = "IP"

    def __init__(self, app, *, arg_subnet, arg_what):
        super().__init__(app)

        if isinstance(arg_what, Expression):
            self.Value = arg_what
        else:
            self.Value = VALUE(app, value=arg_what)

        if isinstance(arg_subnet, Expression):
            self.Subnet = arg_subnet
        else:
            self.Subnet = VALUE(app, value=arg_subnet)

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        subnet = self.Subnet(context, event, *args, **kwargs)

        try:
            if isinstance(subnet, list):
                for element in subnet:
                    if IPAddress(value) in IPNetwork(element):
                        return True
            else:
                return IPAddress(value) in IPNetwork(subnet)

        except netaddr.core.AddrFormatError:
            # IP address could not be detected
            return None

        return False



from ...abc import Expression
from ..value.valueexpr import VALUE
from ...declerror import DeclarationError

import logging

#

L = logging.getLogger(__name__)

#


class DICT(Expression):
    """
    Create or update the dictionary.

    ```
    !DICT
    with: !EVENT
    set:
            item1: foo
            item2: bar
            item3: ...
    unset:
            - item4

    ```

    If `with` is not specified, the new dictionary will be created.
    `add` is also optional.

    This is how to create the empty dictionary:
    ```
    !DICT {}
    ```"""

    Attributes = {
        "With": ["*"],  # TODO: This ...
        "Update": ["*"],  # TODO: This ...
        "Mandatory": [],  # TODO: This ...
        "Unset": [],  # TODO: This ...
    }

    def __init__(
        self,
        app,
        *,
        arg_with=None,
        arg_set=None,
        arg_modify=None,
        arg_unset=None,
        arg_add=None,
        arg_update=None,
        arg_mandatory=None,
        arg_flatten=None
    ):
        super().__init__(app)

        self.With = arg_with

        self.ArgSet = arg_set
        self.ArgFlatten = arg_flatten
        self.ArgModify = arg_modify
        self.ArgAdd = arg_add
        self.Set = None
        self.Flatten = None
        self.Modify = None
        self.Add = None

        self.Unset = arg_unset
        self.Mandatory = arg_mandatory
        self.Update = arg_update

    def set(self, key, value):
        setattr(self, key, value)

        if "Set" in key:
            self.Set[key[3:]] = value
        if "Flatten" in key:
            self.Flatten[key[7:]] = value
        if "Add" in key:
            self.Add[key[3:]] = value
        if "Modify" in key:
            self.Modify[key[6:]] = value

    def initialize(self):
        if self.ArgSet is None:
            self.Set = None
        else:
            assert isinstance(self.ArgSet, dict)
            self.Set = dict()
            self._set_value_or_expression_to_attribute(self.ArgSet, self.Set, "Set")

        if self.ArgFlatten is None:
            self.Flatten = None
        else:
            assert isinstance(self.ArgFlatten, dict)
            self.Flatten = dict()
            self._set_value_or_expression_to_attribute(
                self.ArgFlatten, self.Flatten, "Flatten"
            )

        if self.ArgModify is None:
            self.Modify = None
        else:
            assert isinstance(self.ArgModify, dict)
            self.Modify = dict()
            self._set_value_or_expression_to_attribute(
                self.ArgModify, self.Modify, "Modify"
            )

        if self.ArgAdd is None:
            self.Add = None
        else:
            assert isinstance(self.ArgAdd, dict)
            self.Add = dict()
            self._set_value_or_expression_to_attribute(self.ArgAdd, self.Add, "Add")

    def _set_value_or_expression_to_attribute(self, _from, _to, name):
        for key, value in _from.items():
            # Parse value to attribute, if not expression
            if isinstance(value, Expression):
                _to[key] = value
            else:
                assert (
                    isinstance(value, (int, str, bytes, bool, tuple, list))
                    or value is None
                )
                value = VALUE(self.App, value=value)
                _to[key] = value

            # Set the value to the class attribute
            attr_name = "{}{}".format(name, key)
            setattr(self, attr_name, value)
            self.Attributes[attr_name] = value.get_outlet_type()

    def __call__(self, context, event, *args, **kwargs):
        try:
            if self.With is None:
                with_dict = dict()
            else:
                with_dict = self.With(context, event, *args, **kwargs)
                if with_dict is None:
                    return None
                # TODO: Must be usable as a dictionary

            if self.Set is not None:
                for key, value in self.Set.items():
                    v = value(context, event, *args, **kwargs)
                    if v is not None:
                        with_dict[key] = v

            if self.Flatten is not None:
                for key, value in self.Flatten.items():
                    v = value(context, event, *args, **kwargs)
                    if v is not None and isinstance(v, list):
                        for el in v:
                            with_dict[key + "." + el["Name"]] = el["Value"]
            if self.Modify is not None:
                for key, value in self.Modify.items():
                    # Obtain the original value and pass it to the modify expression
                    orig = with_dict.pop(key, None)

                    modified_value = value(context, event, orig, *args, **kwargs)

                    # Only if modification is successful, store it
                    # Unsetting values should be done via unset
                    if modified_value is not None:
                        with_dict[key] = modified_value

            if self.Add is not None:
                for key, value in self.Add.items():
                    v = value(context, event, *args, **kwargs)
                    if v is not None:
                        try:
                            with_dict[key] += v
                        except KeyError:
                            continue

            if self.Update is not None:
                update_dict = self.Update(context, event, with_dict, *args, **kwargs)
                if update_dict is not None and update_dict is not False:
                    try:
                        with_dict.update(update_dict)
                    except TypeError:
                        pass

            if self.Unset is not None:
                for key in self.Unset:
                    _ = with_dict.pop(key, None)

            # Check that all mandatory fields are present in the dictionary
            if self.Mandatory is not None:
                for mandatory_field in self.Mandatory:
                    if mandatory_field not in with_dict:
                        # TODO: Remove eventually when there are more occurrences among other expressions as well
                        L.warning(
                            "Mandatory field '{}' not present in dictionary. Returning None.".format(
                                mandatory_field
                            )
                        )
                        return None

            return with_dict

        except Exception as e:
            raise DeclarationError(original_exception=e, location=self.get_location())



import os
import tempfile
import logging
import hashlib
import re

from .filesystem import FileSystemLibraryProvider
from ...config import Config

#

L = logging.getLogger(__name__)

#

try:
    import pygit2
except ImportError:
    L.critical(
        "Please install pygit2 package to enable Git Library Provider. >>> pip install pygit2"
    )
    raise SystemExit("Application exiting... .")


class GitLibraryProvider(FileSystemLibraryProvider):
    """
    Read-only git provider to read from remote repository.
    It clones a remote git repository to a temporary directory and then uses the
    FileSystemLibraryProvider to read the files.
    To read from local git repository, please use FileSystemProvider.

    .. code::

            [library]
            providers=git+<URL or deploy token>#<branch name>

            [library:git]
            repodir=<optional location of the repository cache>
    """

    def __init__(self, library, path, layer):
        # format: 'git+http[s]://[<username>:<deploy token>@]<url>[#<branch>]'
        pattern = re.compile(r"git\+(https?://)((.*):(.*)@)?([^#]*)(?:#(.*))?$")
        path_split = pattern.findall(path)[0]
        L.debug(path_split)
        (
            self.URLScheme,
            self.UserInfo,
            self.User,
            self.DeployToken,
            self.URLPath,
            self.Branch,
        ) = path_split
        self.URL = "".join([self.URLScheme, self.UserInfo, self.URLPath])
        self.Branch = self.Branch if self.Branch != "" else None

        repodir = Config.get("library:git", "repodir", fallback=None)
        if repodir is not None:
            self.RepoPath = os.path.abspath(repodir)
        else:
            tempdir = tempfile.gettempdir()
            self.RepoPath = os.path.join(
                tempdir,
                "asab.library.git",
                hashlib.sha256(path.encode("utf-8")).hexdigest(),
            )

        super().__init__(library, self.RepoPath, layer, set_ready=False)

        self.GitRepository = None

        from ...proactor import Module

        self.App.add_module(Module)
        self.ProactorService = self.App.get_service("asab.ProactorService")
        self.PullLock = False

        self.SubscribedPaths = set()

        self.App.TaskService.schedule(self.initialize_git_repository())
        self.App.PubSub.subscribe("Application.tick/60!", self._periodic_pull)

    async def _periodic_pull(self, event_name):
        """
        Changes in remote repository are being pulled every minute. `PullLock` flag ensures that only if previous "pull" has finished, new one can start.
        """
        if self.GitRepository is None:
            return

        if self.PullLock:
            return

        self.PullLock = True

        try:
            to_publish = await self.ProactorService.execute(self._do_pull)
            # Once reset of the head is finished, PubSub message about the change in the subscribed directory gets published.
            for path in to_publish:
                self.App.PubSub.publish("Library.change!", self, path)
        except pygit2.GitError as err:
            L.warning(
                "Periodic pull from the remote repository failed: {}".format(err),
                struct_data={"url": self.URLPath},
            )
        finally:
            self.PullLock = False

    async def initialize_git_repository(self):
        def init_task():
            if pygit2.discover_repository(self.RepoPath) is None:
                # For a new repository, clone the remote bit
                os.makedirs(self.RepoPath, mode=0o700, exist_ok=True)
                self.GitRepository = pygit2.clone_repository(
                    url=self.URL, path=self.RepoPath, checkout_branch=self.Branch
                )
            else:
                # For existing repository, pull the latest changes
                self.GitRepository = pygit2.Repository(self.RepoPath)
                self._do_pull()

        try:
            await self.ProactorService.execute(init_task)

        except KeyError as err:
            pygit_message = str(err).replace('"', "")
            if "'refs/remotes/origin/{}'".format(self.Branch) in pygit_message:
                # branch does not exist
                L.exception(
                    "Branch does not exist.",
                    struct_data={"url": self.URLPath, "branch": self.Branch},
                )
            else:
                L.exception(
                    "Error when initializing git repository: {}".format(pygit_message)
                )
            self.App.stop()  # NOTE: raising Exception doesn't exit the app

        except pygit2.GitError as err:
            pygit_message = str(err).replace('"', "")
            if "unexpected http status code: 404" in pygit_message:
                # repository not found
                L.exception(
                    "Git repository not found.", struct_data={"url": self.URLPath}
                )
            elif "remote authentication required but no callback set" in pygit_message:
                # either repository not found or authentication failed
                L.exception(
                    "Authentication failed when initializing git repository.\n"
                    "Check if the 'providers' option satisfies the format: 'git+<username>:<deploy token>@<URL>#<branch name>'",
                    struct_data={
                        "url": self.URLPath,
                        "username": self.User,
                        "deploy_token": self.DeployToken,
                    },
                )
            elif "cannot redirect from" in pygit_message:
                # bad URL
                L.exception(
                    "Git repository not found.", struct_data={"url": self.URLPath}
                )
            elif "Temporary failure in name resolution" in pygit_message:
                # Internet connection does
                L.exception(
                    "Git repository not initialized: connection failed. Check your network connection.",
                    struct_data={"url": self.URLPath},
                )
            else:
                L.exception("Git repository not initialized: {}".format(err))
            self.App.stop()

        assert hasattr(self.GitRepository, "remotes"), "Git repository not initialized."
        assert (
            self.GitRepository.remotes["origin"] is not None
        ), "Git repository not initialized."
        await self._set_ready()

    def _do_fetch(self):
        """
        It fetches the remote repository and returns the commit ID of the remote branch

        :return: The commit id of the latest commit on the remote repository.
        """
        if self.GitRepository is None:
            return None

        self.GitRepository.remotes["origin"].fetch()
        if self.Branch is None:
            reference = self.GitRepository.lookup_reference("refs/remotes/origin/HEAD")
        else:
            reference = self.GitRepository.lookup_reference(
                "refs/remotes/origin/{}".format(self.Branch)
            )
        commit_id = reference.peel().id
        return commit_id

    def _do_pull(self):
        new_commit_id = self._do_fetch()
        if new_commit_id == self.GitRepository.head.target:
            return []

        # Before new head is set, check the diffs. If changes in subscribed directory occured, add path to "to_publish" list.
        to_publish = []
        for path in self.SubscribedPaths:
            for i in self.GitRepository.diff(
                self.GitRepository.head.target, new_commit_id
            ).deltas:
                if ("/" + i.old_file.path).startswith(path):
                    to_publish.append(path)

        # Reset HEAD
        self.GitRepository.head.set_target(new_commit_id)
        self.GitRepository.reset(new_commit_id, pygit2.GIT_RESET_HARD)

        return to_publish

    async def subscribe(self, path):
        self.SubscribedPaths.add(path)



import collections
import logging

from ..abc.generator import Generator
from .sessionmatrix import SessionMatrix
from .timewindowmatrix import TimeWindowMatrix


L = logging.getLogger(__name__)


class TimeWindowMatrixExportCSVGenerator(Generator):
    def process(self, context, event):
        assert isinstance(event, TimeWindowMatrix)

        def generate(time_window_matrix):
            for i in range(time_window_matrix.Array.shape[0]):
                row_id = time_window_matrix.get_row_name(i)
                if row_id is None:
                    continue

                for j in range(0, time_window_matrix.Dimensions[0]):
                    event = collections.OrderedDict()
                    event["id"] = row_id
                    event["timestamp"] = (
                        time_window_matrix.Start + j * time_window_matrix.Resolution
                    )
                    for k in range(0, time_window_matrix.Dimensions[1]):
                        field_name = "value_{}".format(k)
                        event[field_name] = time_window_matrix.Array[i, j, k]

                    yield event

        return generate(event)


class SessionMatrixExportCSVGenerator(Generator):
    def process(self, context, event):
        assert isinstance(event, SessionMatrix)

        def generate(session_matrix):
            for i in range(0, session_matrix.Array.shape[0]):
                event = collections.OrderedDict()
                row_id = session_matrix.get_row_name(i)
                if row_id is None:
                    continue
                event["id"] = row_id
                for name in session_matrix.Array.dtype.names:
                    if session_matrix.Array.dtype[name].shape == ():
                        event[name] = session_matrix.Array[name][i]
                    else:
                        for j in range(0, session_matrix.Array.dtype[name].shape[0]):
                            for k in range(
                                0, session_matrix.Array.dtype[name].shape[1]
                            ):
                                field_name = "{}_{}_{}".format(name, j, k)
                                if session_matrix.Array.dtype[name].subdtype[
                                    0
                                ].kind in ["f", "u", "i", "b"]:
                                    value = "{0:.10f}".format(
                                        session_matrix.Array[name][i, j, k]
                                    )
                                else:
                                    value = session_matrix.Array[name][i, j, k]

                                event[field_name] = value

                yield event

        return generate(event)



import logging
import json
import yaml
import importlib

from .pipeline import Pipeline

##
L = logging.getLogger(__name__)
##


class PumpBuilder(object):
    """
            PumpBuilder is meant to create the pipeline with connections, processors, sources alternatively.
            ``definition`` is a path to the json file, containing description of the pump.
            Example of such file:

    .. code-block:: json

            {
                    "pipelines" : [
                            {
                                    "id": "MyPipeline0",
                                    "args": {},
                                    "config": {},
                                    "sources": [
                                            {
                                                    "id": "FileCSVSource",
                                                    "module": "bspump.file",
                                                    "class" : "FileCSVSource",
                                                    "args": {},
                                                    "config": {"path":"etc/test.csv", "post":"noop"},
                                                    "trigger": {
                                                            "module": "bspump.trigger",
                                                            "class": "OpportunisticTrigger",
                                                            "id": "",
                                                            "args": {}
                                                    }
                                            }
                                    ],
                                    "processors": [
                                            {
                                                    "module":"bspump-pumpbuilder",
                                                    "class": "Processor00",
                                                    "args": {},
                                                    "config": {}
                                            }
                                    ],
                                    "sink": {
                                            "module":"bspump.common",
                                            "class": "PPrintSink",
                                            "args": {},
                                            "config": {}
                                    }
                            }
                    ]
            }

            |

    """

    def __init__(self, definition):
        """
        Description:

        |

        """
        if isinstance(definition, dict):
            self.Definition = definition
        else:
            with open(definition) as f:
                if definition.endswith(".yaml"):
                    self.Definition = yaml.load(f)
                else:
                    self.Definition = json.load(f)

    def construct_pump(self, app, svc):
        """
                The main method to construct the pump.
                ``app`` is a BspumpApplication object, ``svc``` is service.
                Example of use:

        .. code-block:: python

                app = BSPumpApplication()
                svc = app.get_service("bspump.PumpService")
                pump_builder = PumpBuilder(definition)
                pump_builder.construct_pump(app, svc)
                app.run()

                |

        """
        self.construct_connections(app, svc)
        self.construct_lookups(app, svc)
        self.construct_pipelines(app, svc)

    def construct_connections(self, app, svc):
        """
        Description:

        |

        """

        connections = self.Definition.get("connections")
        if connections is None:
            return

        for i in range(0, len(connections)):
            connection_definition = connections[i]
            self.construct_connection(app, svc, connection_definition)

    def construct_connection(self, app, svc, connection):
        """
        Description:

        |

        """

        module = importlib.import_module(connection["module"])
        connection_class = getattr(module, connection["class"])
        connection_instance = connection_class.construct(app, connection)
        svc.add_connection(connection_instance)

    def construct_lookups(self, app, svc):
        """
        Description:

        |

        """
        lookups = self.Definition.get("lookups")
        if lookups is None:
            return

        for i in range(0, len(lookups)):
            lookup_definition = lookups[i]
            self.construct_lookup(app, svc, lookup_definition)

    def construct_lookup(self, app, svc, lookup):
        """
        Description:

        |

        """
        svc = app.get_service("bspump.PumpService")
        module = importlib.import_module(lookup["module"])
        lookup_class = getattr(module, lookup["class"])
        lookup_instance = lookup_class.construct(app, lookup)
        svc.add_lookup(lookup_instance)

    def construct_pipelines(self, app, svc):
        """
        Description:

        |

        """
        pipelines = self.Definition.get("pipelines")
        if pipelines is None:
            return

        for i in range(0, len(pipelines)):
            pipeline_definition = pipelines[i]
            self.construct_pipeline(app, svc, pipeline_definition)

    def construct_pipeline(self, app, svc, pipeline_definition):
        """
        Description:

        |

        """
        svc = app.get_service("bspump.PumpService")
        pipeline_id = pipeline_definition["id"]
        pipeline = Pipeline(app, pipeline_id)

        sources_definition = pipeline_definition["sources"]
        processors_definition = pipeline_definition.get("processors")
        sink_definition = pipeline_definition["sink"]

        sources = self.construct_sources(app, svc, pipeline, sources_definition)
        pipeline.set_source(sources)
        processors = self.construct_processors(
            app, svc, pipeline, processors_definition
        )
        sink = self.construct_processor(app, svc, pipeline, sink_definition)

        processors.append(sink)
        for processor in processors:
            pipeline.append_processor(processor)

        svc.add_pipeline(pipeline)

    def construct_sources(self, app, svc, pipeline, definition):
        """
        Description:

        |

        """

        sources = []
        for source_definition in definition:
            source = self.construct_source(app, svc, pipeline, source_definition)
            sources.append(source)

        return sources

    def construct_source(self, app, svc, pipeline, definition):
        """
        Description:

        |

        """
        module = importlib.import_module(definition["module"])
        processor_class = getattr(module, definition["class"])
        processor = processor_class.construct(app, pipeline, definition)
        trigger_definition = definition.get("trigger")
        if trigger_definition is None:
            return processor
        trigger = self.construct_trigger(app, svc, trigger_definition)
        return processor.on(trigger)

    def construct_trigger(self, app, svc, definition):
        """
        Description:

        |

        """
        module = importlib.import_module(definition["module"])
        trigger_class = getattr(module, definition["class"])
        trigger = trigger_class.construct(app, definition)
        return trigger

    def construct_processors(self, app, svc, pipeline, definition):
        """
        Description:

        |

        """
        if definition is None:
            return []

        processors = []
        for i in range(0, len(definition)):
            processor_definition = definition[i]
            processor = self.construct_processor(
                app, svc, pipeline, processor_definition
            )
            processors.append(processor)

        return processors

    def construct_processor(self, app, svc, pipeline, definition):
        """
        Description:

        |

        """
        module = importlib.import_module(definition["module"])
        processor_class = getattr(module, definition["class"])
        processor = processor_class.construct(app, pipeline, definition)
        return processor



import os
import time
import json
import asyncio
import logging
import urllib.parse
import configparser

import kazoo.exceptions
import kazoo.recipe.watchers
import kazoo.protocol.states

from .wrapper import KazooWrapper
from ..config import Configurable
from ..log import LOG_NOTICE

#

L = logging.getLogger(__name__)

#


class ZooKeeperContainer(Configurable):
    """
    Create a Zookeeper container with a specifications of the connectivity.
    """

    ConfigDefaults = {
        # Server list to which ZooKeeper Client tries connecting.
        # Specify a comma (,) separated server list.
        # A server is defined as `address:port` format.
        # Example:
        # "servers": "zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181"
        # If servers are empty, default from [zookeeper]/servers will be taken
        # If [zookeeper]/servers value is not provided, ASAB_ZOOKEEPER_SERVERS environment variable is used
        "servers": "",
        # If not provided, "/asab" path will be used
        "path": "",
    }

    def __init__(
        self,
        zookeeper_service,
        config_section_name="zookeeper",
        config=None,
        z_path=None,
    ):
        super().__init__(config_section_name=config_section_name, config=config)

        self.App = zookeeper_service.App
        self.ConfigSectionName = config_section_name
        self.ProactorService = zookeeper_service.ProactorService

        # Parse URL from z_path
        if z_path is not None:
            url_pieces = urllib.parse.urlparse(z_path)
            url_netloc = url_pieces.netloc
            url_path = url_pieces.path
        else:
            url_netloc = ""
            url_path = ""

        # If there is no location, use the value of 'servers' from the configuration
        if url_netloc == "":
            url_netloc = self.Config.get("servers", "")

        if url_netloc == "":
            from ..config import Config

            try:
                url_netloc = Config.get("zookeeper", "servers")
            except (configparser.NoOptionError, configparser.NoSectionError):
                pass

        if url_netloc == "":
            url_netloc = os.environ.get("ASAB_ZOOKEEPER_SERVERS", "")

        if url_netloc == "":
            # if server entry is missing exit
            L.critical(
                "Cannot connect to Zookeeper, the configuration of the server address is not available."
            )
            raise SystemExit("Exit due to a critical configuration error.")

        assert url_netloc is not None

        # If path has not been provided, use the value of 'path' from the configuration
        if url_path == "":
            url_path = self.Config.get("path", "")

        if url_path == "":
            from ..config import Config

            try:
                url_path = Config.get("zookeeper", "path")
            except (configparser.NoOptionError, configparser.NoSectionError):
                pass

        if url_path == "":
            url_path = "asab"

        # Remove all heading '/' from path
        while url_path.startswith("/"):
            url_path = url_path[1:]

        self.Path = url_path

        self.Advertisments = dict()
        self.DataWatchers = set()
        self.App.PubSub.subscribe("Application.tick/300!", self._readvertise)

        self.ZooKeeper = KazooWrapper(self, url_netloc)

        zookeeper_service.Containers.append(self)
        self.ProactorService.schedule(self._start)

    def _start(self):
        # This method is called on proactor thread
        while True:
            try:
                self.ZooKeeper.Client.start(timeout=600)  # 600 seconds
                return
            except Exception as e:
                L.error(
                    "Failed to connect to ZooKeeper: {} (retrying in 2 seconds)".format(
                        e
                    ),
                    struct_data={
                        "hosts": str(self.ZooKeeper.Client.hosts),
                    },
                )
                time.sleep(2)
                continue

    async def _stop(self):
        await self.ZooKeeper._stop()

    def _listener(self, state):
        """
        Generate PubSub events:

        * ZooKeeperContainer.state/CONNECTED!
        * ZooKeeperContainer.state/LOST!
        * ZooKeeperContainer.state/SUSPENDED!
        """
        if state == kazoo.protocol.states.KazooState.CONNECTED:
            self.App.Loop.call_soon_threadsafe(
                self.ZooKeeper.Client.ensure_path, self.Path
            )
            L.log(LOG_NOTICE, "Connected to ZooKeeper")
        else:
            L.warning(
                "ZooKeeper connection state changed", struct_data={"state": str(state)}
            )

        self.App.PubSub.publish_threadsafe(
            "ZooKeeperContainer.state/{}!".format(state), self
        )

    def is_connected(self):
        """
        Check, if the Zookeeper is connected
        """
        return self.ZooKeeper.Client.connected

    # Advertisement into Zookeeper

    def advertise(self, data, path):
        adv = self.Advertisments.get(self.Path + path)
        if adv is None:
            adv = ZooKeeperAdvertisement(self.Path + path)
            self.Advertisments[self.Path + path] = adv
        self.App.TaskService.schedule(adv._set_data(self, data))

    async def _readvertise(self, *args):
        for adv in self.Advertisments.values():
            await adv._readvertise(self)

    # Reading

    async def get_children(self, path=""):
        return await self.ZooKeeper.get_children("{}/{}".format(self.Path, path))

    async def get_data(self, path, encoding="utf-8"):
        raw_data = await self.get_raw_data(path)
        if raw_data is None:
            return {}
        return json.loads(raw_data.decode(encoding))

    async def get_raw_data(self, path):
        return await self.ZooKeeper.get_data("{}/{}".format(self.Path, path))


class ZooKeeperAdvertisement(object):
    def __init__(self, path):
        self.Data = None

        self.Path = path
        self.RealPath = None

        self.Lock = asyncio.Lock()

    async def _set_data(self, zoocontainer, data):
        async with self.Lock:
            if isinstance(data, dict):
                self.Data = json.dumps(data).encode("utf-8")
            elif isinstance(data, str):
                self.Data = data.encode("utf-8")
            else:
                self.Data = data

            def write_to_zk():
                if self.RealPath is None:
                    self.RealPath = zoocontainer.ZooKeeper.Client.create(
                        self.Path,
                        self.Data,
                        sequence=True,
                        ephemeral=True,
                        makepath=True,
                    )
                else:
                    try:
                        zoocontainer.ZooKeeper.Client.set(self.RealPath, self.Data)
                    except kazoo.exceptions.NoNodeError:
                        self.RealPath = zoocontainer.ZooKeeper.Client.create(
                            self.Path,
                            self.Data,
                            sequence=True,
                            ephemeral=True,
                            makepath=True,
                        )

            await zoocontainer.ZooKeeper.ProactorService.execute(write_to_zk)

    async def _readvertise(self, zoocontainer):
        if self.Data is None:
            return

        if self.RealPath is None:
            return

        async with self.Lock:

            def check_at_zk():
                try:
                    if zoocontainer.ZooKeeper.Client.exists(self.RealPath):
                        return

                    # If the advertisement node is not present in the Zookeeper, force the recreation
                    self.RealPath = zoocontainer.ZooKeeper.Client.create(
                        self.Path, self.Data, sequence=True, ephemeral=True
                    )

                except kazoo.exceptions.SessionExpiredError:
                    # The connection to Zookeeper is likely down, this will be retried later
                    pass

            await zoocontainer.ZooKeeper.ProactorService.execute(check_at_zk)



import datetime
import logging

import pytz

import bspump

L = logging.getLogger(__name__)


class TimeZoneNormalizer(bspump.Processor):
    """
    Description: Normalizes datetime from local timezone (e.g. in config) to UTC, which is preferred internal datetime form

    |

    """

    ConfigDefaults = {"timezone": "CET"}

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.TimeZoneSource = pytz.timezone(self.Config.get("timezone"))

    def normalize(self, local_time: datetime.datetime) -> datetime.datetime:
        """
        Description: 	If `local_time` doesn't contain a time zone (e.g. it is naive), the timezone will be added from config

        :return: Normalized `local_time` in UTC

        |

        """
        if not local_time.tzinfo:
            local_time = self.TimeZoneSource.localize(local_time)
        return local_time.astimezone(pytz.utc)

    def process(self, context, event):
        """
        Description: Abstract method to process the event. Must be customized.

        Example:

                >>> native_time = event["@timestamp"]
                >>> local_time = self.normalize(native_time)
                >>> event["@timestamp"] = local_time

        |

        """
        raise NotImplementedError()



import bspump.filter
import bspump.unittest


class AttributeFilterWithAttributeFields(bspump.filter.AttributeFilter):
    def get_fields(self, event):
        return ["attribute"]


class UnpredictableAttributeFilter(bspump.filter.AttributeFilter):
    def predicate(self, event):
        return False


class TestAttributeFilter(bspump.unittest.ProcessorTestCase):
    def test_attribute_filter(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(bspump.filter.AttributeFilter)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])

    def test_test_attribute_filter_inclusive(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(bspump.filter.AttributeFilter, inclusive=True)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{}])

    def test_attribute_filter_with_fields(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(AttributeFilterWithAttributeFields)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{}])

    def test_attribute_filter_with_fields_inclusive(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(AttributeFilterWithAttributeFields, inclusive=True)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])

    def test_unpredictable_attribute_filter(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(UnpredictableAttributeFilter)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])

    def test_unpredictable_attribute_filter_inclusive(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(UnpredictableAttributeFilter, inclusive=True)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])



from .analyzer import AnomalyAnalyzer
from .manager import AnomalyManager
from .storage import AnomalyStorage

from .generalanomaly import GeneralAnomaly

__all__ = (
    "AnomalyAnalyzer",
    "AnomalyManager",
    "AnomalyStorage",
    "GeneralAnomaly",
)



from jinja2 import Environment, FileSystemLoader
import importlib.util
from pathlib import Path
import os


def find_module_path(module_name):
    spec = importlib.util.find_spec(module_name)
    if spec and spec.origin:
        return spec.origin
    else:
        return f"Module '{module_name}' not found or built-in."


path = str(
    Path(os.path.split(find_module_path("bspump"))[:-1][0])
    / "http"
    / "web"
    / "templates"
)
env = Environment(loader=FileSystemLoader(path))



import operator

from ..abc import SequenceExpression, Expression
from ..declerror import DeclarationError
from .value.valueexpr import VALUE
from .datastructs.itemexpr import ITEM, ITEM_optimized_EVENT_VALUE

from .value.eventexpr import EVENT
from .utility.context import CONTEXT


class ComparisonExpression(SequenceExpression):
    Category = "Compare"

    def __call__(self, context, event, *args, **kwargs):
        it = iter(self.Items)

        i = next(it)
        try:
            a = i(context, event, *args, **kwargs)
        except Exception as e:
            raise DeclarationError(original_exception=e, location=i.get_location())

        for i in it:
            try:
                b = i(context, event, *args, **kwargs)
            except Exception as e:
                raise DeclarationError(original_exception=e, location=i.get_location())

            if not self.Operator(a, b):
                return False
            a = b

        return True

    def get_outlet_type(self):
        return bool.__name__


class LT(ComparisonExpression):
    """
    Operator '<'
    """

    Operator = operator.lt


class LE(ComparisonExpression):
    """
    Operator '<='
    """

    Operator = operator.le


class EQ(ComparisonExpression):
    """
    Operator '=='
    """

    Operator = operator.eq

    Attributes = {
        "Items": [
            "si64",
            "si8",
            "si16",
            "si32",
            "si64",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
            "str",
        ]
    }

    def optimize(self):
        if len(self.Items) == 2 and isinstance(self.Items[1], VALUE):
            if isinstance(self.Items[0], ITEM_optimized_EVENT_VALUE):
                return EQ_optimized_EVENT_VALUE(self)

            # The nested objects may not be optimized yet when this parent optimization is called
            if isinstance(self.Items[0], ITEM):
                if isinstance(self.Items[0].With, EVENT) and isinstance(
                    self.Items[0].Item, VALUE
                ):
                    return EQ_optimized_EVENT_VALUE(self)
                elif (
                    isinstance(self.Items[0].With, CONTEXT)
                    and isinstance(self.Items[0].Item, VALUE)
                    and "." in self.Items[0].Item.Value
                ):
                    # Skip nested context from optimization
                    return None
                else:
                    return EQ_optimized_simple(self)

        return None

    def get_items_inlet_type(self):
        # Find the first usable type in the items
        for item in self.Items:
            outlet_type = item.get_outlet_type()
            if outlet_type not in frozenset(["^"]):
                return outlet_type
        raise NotImplementedError("Cannot decide on items inlet type '{}'".format(self))

    def consult_inlet_type(self, key, child):
        return self.get_items_inlet_type()


class EQ_optimized_simple(EQ):
    def __init__(self, orig):
        super().__init__(orig.App, sequence=orig.Items)
        self.A = self.Items[0]
        assert isinstance(self.A, Expression)

        assert isinstance(self.Items[1], VALUE)
        self.B = self.Items[1].Value
        assert isinstance(self.B, (bool, str, int, float))

    def __call__(self, context, event, *args, **kwargs):
        return self.A(context, event, *args, **kwargs) == self.B

    def optimize(self):
        return None


class EQ_optimized_EVENT_VALUE(EQ):
    def __init__(self, orig):
        super().__init__(orig.App, sequence=orig.Items)
        self.Akey = self.Items[0].Item.Value
        self.Adefault = self.Items[0].Default.Value

        assert isinstance(self.Items[1], VALUE)
        self.B = self.Items[1].Value
        assert isinstance(self.B, (bool, str, int, float))

    def __call__(self, context, event, *args, **kwargs):
        return event.get(self.Akey, self.Adefault) == self.B

    def optimize(self):
        return None


class NE(ComparisonExpression):
    """
    Operator '!='
    """

    Operator = operator.ne


class GE(ComparisonExpression):
    """
    Operator '>='
    """

    Operator = operator.ge


class GT(ComparisonExpression):
    """
    Operator '>'
    """

    Operator = operator.gt

    def get_items_inlet_type(self):
        return evaluate_items_inlet_type(self.Items)


# TODO: This operator is obsoleted and should be removed (AT Jan 2021)
class IS(ComparisonExpression):
    """
    Operator 'is'
    """

    Operator = operator.is_


# TODO: This operator is obsoleted and should be removed (AT Jan 2021)
class ISNOT(ComparisonExpression):
    """
    Operator 'is not'
    """

    Operator = operator.is_not

    def get_items_inlet_type(self):
        # Find the first usable type in the items
        for item in self.Items:
            outlet_type = item.get_outlet_type()
            if outlet_type not in frozenset(["^"]):
                return outlet_type
        raise NotImplementedError("Cannot decide on items inlet type '{}'".format(self))

    def consult_inlet_type(self, key, child):
        return self.get_items_inlet_type()


def evaluate_items_inlet_type(items):
    strings = False
    signed_integers = False
    unsigned_integers = False
    max_integer_size = 0

    for item in items:
        outlet_type = item.get_outlet_type()
        if outlet_type == ["^"]:
            continue

        iti = integer_type_info.get(outlet_type)
        if iti is not None:
            if max_integer_size < iti:
                max_integer_size = iti
            if outlet_type[:1] == "u":
                unsigned_integers = True
            else:
                signed_integers = True
            continue

        raise NotImplementedError(
            "Unsupported type '{}' in comparison operator".format(outlet_type)
        )

    if strings is False and unsigned_integers is False and signed_integers is True:
        # Just signed integers found
        return "si{}".format(max_integer_size)

    if strings is False and signed_integers is False and unsigned_integers is True:
        # Just unsigned integers found
        return "ui{}".format(max_integer_size)

    if strings is False and signed_integers is True and unsigned_integers is True:
        if max_integer_size == 1:
            # Booleans are handed differently (this is weird combo anyway)
            return "ui1"
        # Mix of signed and unsigned integers found
        return "si{}".format(max_integer_size << 1)

    return "?"


integer_type_info = {
    "ui1": 1,
    "si8": 8,
    "si16": 16,
    "si32": 32,
    "si64": 64,
    "si128": 128,
    "si256": 256,
    "ui8": 8,
    "ui16": 16,
    "ui32": 32,
    "ui64": 64,
    "ui128": 128,
    "ui256": 256,
}
